// Component TBASSPlayer
//
//    written by Silhwan Hyun  hsh@chollian.net
//
//
// Revision history
//
// (Note) visualizer = Winamp visualization plug-in or its visible window
//
//
// Ver 1.51                       24 Jan 2004
//  Introduced sub-component TVisThread to raise refresh rate of visualizer.
//   (The maximum refresh rate is changed from 30 to 100 fps.)
//  VisOut.exe is launched by internal code of BASSPlayer.pas.
//   (Appexec.pas is not needed to launch VisOut.exe.)
//  Inhibited multiple instance of VisOut.exe because they will use same address
//   space obtained by fisSharedMemory1.
//

// Ver 1.5                        15 Dec 2003
//   ( Based on bass.dll/basswma.dll Ver 2.0 )
//
//  Added properties to play stream files on the internet, or the stream from internet
//  radio (Shout/Icecast server, WMA broadcaster).
//  Followings are the new properties for them.
//    - IsNetStream : true if the stream file is on the internet
//    - IsNetRadio  : true if the stream comes from Shout/Icecast server
//    - ICYInfo     : pointer to ICY tag
//    - HTTPInfo    : pointer to HTTP tag
//    - DownloadProgress : download progress in bytes of an internet file stream
//    - OnGetMeta   : event handler for catching Meta data from Shout/Icecast server.
//    - OnGetChannelInfo : event handler for catching information of an internet stream,
//                         obtained from Winamp input plug-ins.
//    - OnDownloaded : event handler for catching re-estimated value of the playback length
//                      & tag data of an internet stream.
//  Added property "DecodingByPlugin" to show which takes charge of decoding stream,
//     (true if it is Winamp input plug-in, false if BASS)
//  Added property "Seekable" to show if playback position can be changed.
//  Added a function and a procedure to support Winamp DSP plug-ins.
//    - function : RunDSPPlugin         // Run a DSP plug-in
//    - procedure : QuitDSPPlugin       // Quit a DSP plug-in
//  note) If you want to activate Winamp DSP plug-in, you must use appropriate Winamp
//        input plug-in for the decoder of stream file. (ex. load in_mp3.dll and set
//        'Plugin First' check box using 'Input plug-in Configuration' form, if you
//        are playing MP3 stream files)
//
//  Added a function to inform player status to visualizer
//    - procedure : InformPlayerStatus  // Inform player status to visualizer
//
//  Added some parameters at running VisOut.exe for more proper operation.
//
//  Changed sound volume range : 0 ~ 100 -> 0 ~ 255
//

// Ver 1.4                        12 May 2003
//   ( Based on bass.dll/basswma.dll Ver 1.8 )
//  Added(modified) functions, procedures and properties to set up synchronized action
//   between the windows of main program and visualizer.
//   - New function : GetVisWindowHandle     // get the handle of visualizer
//   - Modified function : RunVisoutProgram  // added a parameter to inform visualizer
//                                           //  the window handle of main program
//   - New procedure : HideVisWindow         // Hide visualizer
//                     ShowVisWindow         // Show visualizer
//   - New property : VisScale               // set(get) output level of FFT (min : 256 ~ max : 1024)
//                    SyncVisWindow          // set(get) status of synchronized action (true : enabled)
//

// Ver 1.3                        5 Apr 2003
//   ( Based on bass.dll/basswma.dll Ver 1.8 )
//  Added a function and a procedure to run Winamp Visualization plug-ins
//   - function RunVisoutProgram(VismodName : string; VismodNo : word) : boolean;
//   - procedure QuitVisoutProgram;
//  Modified function GetVolume and procedure SetVolume to reflect/adjust the
//   volume level of Winamp plug-in's output
//  Added property EchoLevel & ReverbLevel to adjust the intensity of sound effect
//

// Ver 1.2                        5 Feb 2003
//   ( Based on bass.dll/basswma.dll Ver 1.7 )
//  Added functions & properties to use Winamp 2 input plug-ins.
//   Now TBASSPlayer can play stream files which BASS sound system cannot decode
//   if provided appropriate Winamp input plug-in.
//  Changed the loading method of bass.dll & basswma.dll from static loading to
//   dynamic loading
//
//   - Modified function : GetStreamInfo
//   - New properties : BASSWMAReady, Version, NativeFileExts, PluginFileExts, CanUsePlugin,
//                      SupportedBy, PluginFirst, OnPluginStartPlay, UnloadButEnabled
//   - Renamed property : DLLVersionStr -> BASSDLLVer
//   - New  procedure : ShowPluginConfigForm (-> used to show PluginConfigForm,
//                        which is defined in sub unit PluginConfig.pas)
//   - New functions in sub unit PluginCtrl which you can call in main program
//     if you want to manage Winamp input plug-ins directly(= not using PluginConfigForm).
//       GetPluginIndex(PluginName : string) : integer;
//       GetPlugin(PluginNum : integer) : TPlugin;
//       GetPluginInfo(PluginNum : integer) : TPluginInfo;
//       LoadWinampPlugin(PluginName : string) : integer;
//       UnloadWinampPlugin(PluginName : string) : integer;
//  * include "PluginCtrl" in uses clause of main program to use above functions
//
//  * The items which does not affect external program are not mentioned here
//

// Ver 1.1                         8 Dec 2002
//   ( Based on bass.dll/basswma.dll Ver 1.7 )
//   - New function : GetChannelFFTData (for custom processing of FFT data)
//   - Added field at record TStreamInfo : Year
//   - Added property : EQBands (for custom setting of equalizer bands)
//   - Renamed property : EQGains ( <- EqPreset )
//   - New type identifiers : TBandData, TEQBands
//   - Renamed type identifier : TEQGains ( <- TEqPreset )
//   - Updated procedure : TimerFFTTimer (for better output of spectrum visualization)
//

// Ver 1.0                         4 Aug 2002
//   ( Based on bass.dll/basswma.dll Ver 1.6 )
//   - Initial release

unit BASSPlayer;

interface

uses
  Windows, Messages, SysUtils, Classes, Forms, Controls, StdCtrls, ExtCtrls,
  Dynamic_BASS, RT_BASSWMA, ioplug, PluginCtrl, MPEGAudio, OggVorbis, WMAFile, WAVFile,
  PluginConfig, fisSharedMemory, Dialogs;

const
   Ver = '1.5';
   MaxVolume = 255;      // Volume range : 0 ~ 255
   MaxDX8Effect = 32;    // DX8 Effect range : 0 ~ 32
   NumFFTBands = 25;     // Number of bands for spectrum visualization
   NumEQBands  = 10;     // Number of bands for equalizer
   EQBandWidth  : array[0..NumEQBands-1] of float =
                  (4, 4, 4, 4, 5, 6, 5, 4, 3, 3);
   EQFreq : array[0..NumEQBands-1] of float =
                  (80, 170, 310, 600, 1000, 3000, 6000, 10000, 12000, 14000);
   NoDX8Msg : pchar = 'DirectX Ver 8 are not installed.' + #13#10 +
                      'Sound effects except flanger are disabled.';

   WM_GetToEnd      = WM_USER + 107;  // message to notify that BASS reaches the end of a stream
   WM_NewFFTData    = WM_USER + 108;  // message to notify that new FFT data are ready
   WM_GetMeta       = WM_USER + 110;  // message to notify that new Shoutcast Metadata are ready
   WM_DownLoaded    = WM_USER + 111;  // message to notify that downloading of an internet stream is done.

type
  TSoundEffects = set of (Flanger, Equalizer, Echo, Reverb);
  TFFTData = array [0..512] of Single;
  TBandOut = array[0..NumFFTBands-1] of word;
  TChannelType = (Channel_NotOpened, Channel_Stream, Channel_WMA, Channel_Music, Channel_Plugin);
  TSupportedBy = (Both, BASSNative, WinampPlugin, None);
  TEQGains = array [0..NumEQBands-1] of float;
  TBandData = record
     CenterFreq : float;
     Bandwidth  : float;
  end;
  TEQBands = record
     Bands : word;     // Number of equalizer bands (min : 0, max : 10)
     BandPara : array[0..NumEQBands-1] of TBandData;
  end;

  TNotifyEvent = procedure(Sender: TObject) of object;
  TNotifyNetEvent = procedure(Sender: TObject; Content : string) of object;
  TNotifyPluginEvent = procedure(Sender: TObject; ChannelInfo : PChannelInfo) of object;
  TNotifyNewFFTDataEvent = procedure(Sender: TObject; BandOut : TBandOut) of object;

  TVisThread = class(TComponent)   // * New at Ver 1.51
  private
    { Private declarations }
    FSuspended : boolean;
    FDelayMS : DWORD;
    FLatencyMs : DWORD;
    FChannelId : DWORD;
    FSampleRate : DWORD;
    FChannels : DWORD;
    FPosition : DWORD;
    FVisScale : word;

    WaitCounter : integer;
    Timer1: TTimer;
    procedure Timer1Timer(Sender: TObject);
  public
    { Public declarations }
    fisSharedMemory1: TfisSharedMemory;
    ShareMemPointer : pointer;
    DataReadyMsg : hwnd;

    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    procedure SetDelayMS(DelayMS : DWORD);
    procedure SetChannelInfo(ChannelId, SampleRate, Channels : DWORD);
    procedure SetPosition(Position : DWORD);
    procedure SetVisScale(VisScale : word);
    function  Suspended : boolean;
    procedure Suspend;
    procedure Resume;
  end;

  TBASSPlayer = class(TComponent)
  private
    { Private declarations }
    AppHWND     : THandle;
    TimerFFT    : TTimer;

    ChannelType   : TChannelType;
    Channel       : DWORD;

    BassChannelInfo : BASS_CHANNELINFO;
    MessageHandle : hwnd;

    APlugin : TPlugin;
    MessageHandle2 : hwnd;
    NowStarting : boolean;
    UsesOutputPlugin : boolean;
 // Some input plug-ins do not need output plug-in. i.e., some input plug-ins
 // sound out decoded data directly.
 // UsesOutputPlugin is used to know if input plug-in needs output plug-in.
 // Use this instead of input plug-in's UsesOutPlug property because some
 // input plug-ins report wrong UsesOutPlug value.

    BASSDLLLoaded : boolean;
    BassInfoParam : BASS_INFO;
    EQParam       : BASS_FXPARAMEQ;
    EchoParam     : BASS_FXECHO;
    ReverbParam   : BASS_FXREVERB;
    FSoundEffects : TSoundEffects;

    DSPHandle     : HDSP;
    fladspHandle  : HDSP;
    EQHandle      : array [0..NumEQBands-1] of HFX;
    EchoHandle    : HFX;
    ReverbHandle  : HFX;

    BandOut: TBandOut;

    NetStream  : boolean;
    NetRadio   : boolean;
    ICYTag     : pchar;
    HTTPTag    : pchar;

    MPEG        : TMPEGaudio;
    Vorbis      : TOggVorbis;
    WMA         : TWMAfile;
    WAV         : TWAVFile;

  // variables for visualizer
    VisProcId : DWORD;
    VisWindowHandle : HWND;
    GoVisOut : boolean;
    FVisScale : word;
    FSyncVisWindow : boolean;
    WaitingVisualizer : boolean;

    PluginConfigForm: TPluginConfigForm;
    VisThread: TVisThread;

    FBASSReady      : boolean;
    FBASSWMAReady   : boolean;
    FPluginNum      : integer;
    FPluginFirst    : boolean;
    FVersionStr     : string;
    FDLLVersionStr  : string;
    FStreamName     : string;
    FDecodingByPlugin : boolean;
    FDownLoaded     : boolean;
    FStreamInfo     : TStreamInfo;

    FOutputVolume   : DWORD;
    FDX8EffectReady : boolean;
    FEchoLevel      : word;
    FReverbLevel    : word;
    FEQGains        : TEQGains;
    FEQBands        : TEQBands;
    FPlayerMode     : TPlayerMode;
    FUnloadButEnabled : boolean;

    FOnPlayEnd         : TNotifyEvent;
    FOnPluginStartPlay : TNotifyPluginEvent;
    FOnGetChannelInfo  : TNotifyPluginEvent;
    FOnGetMeta         : TNotifyNetEvent;
    FOnDownloaded      : TNotifyNetEvent;
    FOnNewFFTData      : TNotifyNewFFTDataEvent;

    procedure ClearEffectHandle;
    function  IsPluginReady : boolean;
    function  GetStreamInfo2(StreamName : string;
                            var StreamInfo : TStreamInfo;
                            var SupportedBy : TSupportedBy;
                            var PluginNum : integer) : boolean;
    procedure SetVersionStr(Value : string);
    procedure SetDLLVersionStr(Value : string);
    procedure SetDX8EffectReady(Value : boolean);
    procedure SetVolume(Value : DWORD);
    procedure SetEchoLevel(Value : word);
    procedure SetReverbLevel(Value : word);
    procedure SetSoundEffect(Value : TSoundEffects);
    procedure SetEQGains(Value : TEQGains);
    procedure SetEQBands(Value : TEQBands);
    procedure SetPluginFirst(Value : boolean);
    procedure SetUnloadBut(Value : boolean);
    procedure SetSyncVisWindow(Value : boolean);  // * New at Ver 1.4
    procedure SetVisScale(Value : word);
    function  GetDLLVersionStr : string;
    function  GetNativeFileExts : string;
    function  GetPluginFileExts : string;
    function  IsSeekable : boolean;               // * New at Ver 1.5
    function  GetDownloadProgress : DWORD;        // * New at Ver 1.5
    procedure PausePlay;
    procedure ResumePlay;
    procedure Restart;
    procedure TimerFFTTimer(Sender: TObject);
    procedure ProcMessage(var Msg: TMessage);
    function  CurrentPosition : DWORD;
    procedure SetPosition(Value : DWORD);
    procedure InformPlayerStatus(StatusIndex : DWORD);   // * New at Ver 1.5
    procedure Error(msg: string);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    function  GetVisWindowHandle : HWND;        // * New at Ver 1.4
    function  GetStreamInfo(StreamName : string;
                            var StreamInfo : TStreamInfo;
                            var SupportedBy : TSupportedBy) : boolean;

    function  Open(StreamName : string) : boolean;
    function  PlayLength : DWORD;
    function  GetChannelFFTData(PFFTData : pointer; FFTFlag : DWORD) : boolean;

    procedure Play;
    procedure Stop;
    procedure Pause(pAction : boolean);  // true : pause, false : resume
    procedure Close;

    procedure ShowPluginConfigForm;
    property  EQGains : TEQGains read FEQGains write SetEQGains;
    property  EQBands : TEQBands read FEQBands write SetEQBands;
    property  PlayerReady : boolean read FBASSReady;
    property  BASSWMAReady : boolean read FBASSWMAReady;
 //   property  SupportedBy : TSupportedBy read FSupportedBy;     // * Obsolete at Ver 1.5
    property  DecodingByPlugin : boolean read FDecodingByPlugin;  // * New at Ver 1.5
    property  Seekable : boolean read IsSeekable;  // * New at Ver 1.5
    property  StreamInfo : TStreamInfo read FStreamInfo;
    property  Mode : TPlayerMode read FPlayerMode;
    property  Position : DWORD read CurrentPosition write SetPosition;
    property  StreamName : string read FStreamName;

    property  ChannelId : DWORD read Channel;
    property  Volume : DWORD read FOutputVolume write SetVolume;  // 0(minimum output) ~ MaxVolume(maximum output)
    property  SoundEffects : TSoundEffects read FSoundEffects write SetSoundEffect;

  // EchoLevel & ReverbLevel : 0(minimum sound effect) ~ MaxDX8Effect(maximum sound effect)
    property  EchoLevel : word read FEchoLevel write SetEchoLevel;
    property  ReverbLevel : word read FReverbLevel write SetReverbLevel;

    property  CanUsePlugin : boolean read IsPluginReady;
    property  NativeFileExts : string read GetNativeFileExts;
    property  PluginFileExts : string read GetPluginFileExts;
    property  UnloadButEnabled : boolean read FUnloadButEnabled write SetUnloadBut;

    property  IsNetStream : boolean read NetStream;  // * New at Ver 1.5
    property  IsNetRadio : boolean read NetRadio;    // * New at Ver 1.5
    property  ICYInfo : pchar read ICYTag;           // * New at Ver 1.5
    property  HTTPInfo : pchar read HTTPTag;         // * New at Ver 1.5
    property  DownloadProgress : DWORD read GetDownloadProgress; // * New at Ver 1.5
    function  RunVisoutProgram(VismodName : string; VismodNo : word;
                              MainWinHwnd : HWND) : boolean; // * Modified at Ver 1.5
    procedure QuitVisoutProgram;       // * New at Ver 1.3
    function  RunDSPPlugin(DSPmodName : string; DSPmodNo : word;
                              MainWinHwnd : HWND) : boolean; // * New at Ver 1.5
    procedure QuitDSPPlugin;           // * New at Ver 1.5
    procedure HideVisWindow;           // * New at Ver 1.4
    procedure ShowVisWindow;           // * New at Ver 1.4

    function  GetMessageHandle : HWND;   // debug purpose (to debug VisOut program)
  published
    { Published declarations }
    property  Version : string read FVersionStr write SetVersionStr;
    property  BASSDLLVer : string read FDLLVersionStr write SetDLLVersionStr;
    property  DX8EffectReady : boolean read FDX8EffectReady write SetDX8EffectReady;
    property  PluginFirst : boolean read FPluginFirst write SetPluginFirst;
    property  OnNewFFTData : TNotifyNewFFTDataEvent read FOnNewFFTData write FOnNewFFTData;
    property  VisScale : word read FVisScale write SetVisScale;
    property  SyncVisWindow : boolean read FSyncVisWindow write SetSyncVisWindow; // * New at Ver 1.4
    property  OnPluginStartPlay : TNotifyPluginEvent read FOnPluginStartPlay
                                           write FOnPluginStartPlay;
    property  OnGetChannelInfo : TNotifyPluginEvent read FOnGetChannelInfo   // * New at Ver 1.5
                                           write FOnGetChannelInfo;
    property  OnGetMeta : TNotifyNetEvent read FOnGetMeta write FOnGetMeta;  // * New at Ver 1.5
    property  OnDownloaded : TNotifyNetEvent read FOnDownloaded write FOnDownloaded; // * New at Ver 1.5
    property  OnPlayEnd : TNotifyEvent read FOnPlayEnd write FOnPlayEnd;
  end;

  procedure Register;


implementation


procedure WinProcessMessages;
// Allow Windows to process other system messages
var
    ProcMsg  :  TMsg;
begin
    while PeekMessage(ProcMsg, 0, 0, 0, PM_REMOVE) do begin
      if (ProcMsg.Message = WM_QUIT) then Exit;
      TranslateMessage(ProcMsg);
      DispatchMessage(ProcMsg);
    end;
end;


constructor TVisThread.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);

   Timer1 := TTimer.Create(nil);
   Timer1.Interval := 10;
   Timer1.Enabled := false;
   Timer1.OnTimer := Timer1Timer;

   fisSharedMemory1 := TfisSharedMemory.Create(Self);
   fisSharedMemory1.ShareName := 'VisOutSharedMemory';  // * Changed at Ver 1.51
   fisSharedMemory1.Size := 2600;                       // * Changed at Ver 1.51
   fisSharedMemory1.TargetApplication := 'VisOut';
   if fisSharedMemory1.CreateMemory then
      ShareMemPointer := fisSharedMemory1.MapMemory
   else
      ShareMemPointer := nil;
   DataReadyMsg := fisSharedMemory1.SendingDataMsg;

   FSuspended := true;
end;

destructor  TVisThread.Destroy;
begin
   Timer1.Free;

   if ShareMemPointer <> nil then
      fisSharedMemory1.UnMapMemory(ShareMemPointer);
   if fisSharedMemory1 <> nil then
      fisSharedMemory1.CloseMemory;
   fisSharedMemory1.Free;

   inherited Destroy;
end;

procedure TVisThread.SetDelayMS(DelayMS : DWORD);
begin
   FLatencyMS := (DelayMS shr 16);
   FDelayMS := (DelayMS and $0000ffff);
   if FDelayMS < 10 then
      FDelayMS := 10;
   Timer1.Interval := FDelayMS;
end;

procedure TVisThread.SetChannelInfo(ChannelId, SampleRate, Channels : DWORD);
begin
   FChannelId := ChannelId;
   FSampleRate := SampleRate;
   FChannels := Channels;
end;

procedure TVisThread.SetPosition(Position : DWORD);
begin
   FPosition := Position;
end;

procedure TVisThread.SetVisScale(VisScale : word);
begin
   FVisScale := VisScale;
end;


function  TVisThread.Suspended : boolean;
begin
   result := FSuspended;
end;

procedure TVisThread.Suspend;
begin
   FSuspended := true;
   Timer1.Enabled := false;
end;

procedure TVisThread.Resume;
var
   p1 : PDWORD;
begin
   Fsuspended := false;
   WaitCounter := 0;
   p1 := ShareMemPointer;
   inc(p1, 280);
   p1^ := 0;   // Clear flag which is used to notify that new data is given
   Timer1.Enabled := true;
end;


procedure TVisThread.Timer1Timer(Sender: TObject);
var
   i : integer;
   p1 : PDWORD;
   p2 : PBYTE;
   BytesValid : longint;
   SpectrumData: array[0..4097] of float;
   WaveformData: array[0..2303] of smallint;
begin
   Timer1.Enabled := false;
   if FSuspended then
      exit;

   if BASS_ChannelIsActive(FChannelId) <> BASS_ACTIVE_PLAYING then
   begin
      Timer1.Enabled := true;
      exit;
   end;

   BytesValid := Bass_ChannelGetdata(FChannelId, nil, BASS_DATA_AVAILABLE);
   if BytesValid < (2048 * FChannels) then
   begin
      Timer1.Enabled := true;
      exit;
   end;

   p1 := ShareMemPointer;
   inc(p1, 280); // FFT Data region is allocated from offset 280 thru 2599 in shared memory

 // The DWORD of offset 280 is used as a flag to notify that new data is given.
 // It is to be zero if rendering is completed.
 // But it won't be zero in some cases so I use WaitCounter to escape from such situation.
   if p1^ <> 0 then
   begin
      WaitCounter := WaitCounter + 1;
      if WaitCounter < 10 then
      begin
         Timer1.Enabled := true;
         exit;
      end;
   end;
   WaitCounter := 0;

   p1^ := 1;     // Set flag to notify that new data is given
   inc(p1, 1);
   p1^ := FSampleRate;
   inc(p1, 1);
   p1^ := FChannels;
   inc(p1, 1);
   p1^ := FPosition;
   inc(p1, 1);
   p2 := pointer(p1);

   if FChannels = 1 then
   begin
      Bass_ChannelGetdata(FChannelId, @spectrumdata, BASS_DATA_FFT2048);

      for i := 1 To 576 do
      begin
        p2^ := trunc(SpectrumData[i] * FVisScale);
        inc(p2);
      end;

      for i := 1 To 576 do
      begin
        p2^ := trunc(SpectrumData[i] * FVisScale);
        inc(p2);
      end;
   end else

// Add BASS_DATA_FFT_INDIVIDUAL flag for seperate FFT per each channel
   begin
      Bass_ChannelGetdata(FChannelId, @spectrumdata, BASS_DATA_FFT2048 + BASS_DATA_FFT_INDIVIDUAL);

      for i := 1 To 576 do
      begin
        p2^ := trunc(SpectrumData[i*2] * FVisScale);
        inc(p2);
      end;

      for i := 1 To 576 do
      begin
        p2^ := trunc(SpectrumData[i*2 + 1] * FVisScale);
        inc(p2);
      end;
   end;

   if FChannels = 1 then
   begin
     Bass_ChannelGetdata(FChannelId, @waveformdata, 1152);

     for i := 0 to 575 do
     begin
       p2^ := waveformdata[i] div 256;
       inc(p2);
     end;

     for i := 0 to 575 do
     begin
       p2^ := waveformdata[i] div 256;
       inc(p2);
     end;
   end else
   begin
      Bass_ChannelGetdata(FChannelId, @waveformdata, 2304);

      for i := 0 to 575 do
      begin
        p2^ := waveformdata[i*2] div 256;
        inc(p2);
      end;

      for i := 0 to 575 do
      begin
        p2^ := waveformdata[i*2+1] div 256;
        inc(p2);
      end;
   end;

 // Inform "VisOut.exe" that visualization data is ready
   fisSharedMemory1.InformDataReady(10{DataReady}, 2320);
   if not FSuspended then
      Timer1.Enabled := true;
end;

// Following statements for flanger effect is adopted from DTMain.pas of BASS16
// package

const
   FLABUFLEN = 350;         // buffer length for flanger effect

var
   // Variables for DSP (flanger effect) implementation
   flabuf : array[0..FLABUFLEN-1, 0..2] of SmallInt;  // buffer
   flapos : Integer;         // cur.pos
   flas, flasinc : FLOAT;    // sweep pos/min/max/inc

function fmod(a, b: FLOAT): FLOAT;
begin
   Result := a - (b * Trunc(a / b));
end;

function Clip(a: Integer): Integer;
begin
   if a <= -32768 then
      a := -32768
   else if a >= 32767 then
      a := 32767;

   Result := a;
end;

procedure Flange(handle: HSYNC; channel: DWORD; buffer: Pointer; length, user: DWORD); stdcall;
var
  lc, rc: SmallInt;
  p1, p2, s: Integer;
  d: ^DWORD;
  f: FLOAT;
begin
  d := buffer;
  while (length > 0) do
  begin
    lc := LOWORD(d^); rc := HIWORD(d^);
    p1 := (flapos + Trunc(flas)) mod FLABUFLEN;
    p2 := (p1 + 1) mod FLABUFLEN;
    f := fmod(flas, 1.0);
    s := lc + Trunc(((1.0-f) * flabuf[p1, 0]) + (f * flabuf[p2, 0]));
    flabuf[flapos, 0] := lc;
    lc := Clip(s);
    s := rc + Trunc(((1.0-f) * flabuf[p1, 1]) + (f * flabuf[p2, 1]));
    flabuf[flapos, 1] := rc;
    rc := Clip(s);
    d^ := MakeLong(lc, rc);
    Inc(d);
    Inc(flapos);
    if (flapos = FLABUFLEN) then flapos := 0;
    flas := flas + flasinc;
    if (flas < 0) or (flas > FLABUFLEN) then
      flasinc := -flasinc;
    length := length - 4;
  end;
end;

const
  decblock = 4608;
var
  DSPmod : TDSPmod;
  DSPModNum : word;

  exlen : integer = 0;
//  Excess : array[0..decblock * 2 - 1] of byte;

// Following WinampDSPProc procedure is not completed and not used in TBassPlayer.
// Anyway you can use any Winamp DSP plug-ins by direct coupling of input~DSP plug-ins.
// See PluginCtrl.pas if you want to know the method for direct coupling of input~DSP plug-ins.
// The direct coupling of input~DSP plug-ins is simple and reliable, so I leave this
// WinampDSPProc procedure uncompleted.
procedure WinampDSPProc(handle: HSYNC; channel: DWORD; buffer: Pointer; length, user: DWORD); stdcall;
var
//  DSPbuffer : array[0..4607] of byte;

  ReadP  : pointer;
  c      : DWord;
  a      : Integer;
//  numsamples : integer;
begin
   a := length;
   ReadP := buffer;
   while a > 0 do
   begin
      if a > (576 * 4) then
         c := (576 * 4)
      else
         c := a;
    //  Move(ReadP^, DSPbuffer, c);
    //  numsamples := DSPmod[DSPModNum].ModifySamples(DSPmod[DSPModNum], @DSPbuffer, c Div 4, 16, 2, 44100);
    //  Move(DSPbuffer, ReadP^, c);
      a := a - c;
      if a > 0 then
         DWord(ReadP) := DWord(ReadP) + c;
   end;
end;

var
   TitleP : array[0..255] of char;

// This procedure is called when metadata are received in a Shoutcast stream.
procedure MetaSync(SyncHandle : HSYNC; Channel, data, user : DWORD); stdcall;
var
   TagP : pchar;
   tmpStr, StreamTitle : string;
   TitlePos, DelimeterPos : integer;
begin
   StreamTitle := '';
   TagP := pchar(data);
   tmpStr := string(TagP);
   TitlePos := Pos('StreamTitle=', tmpStr);
   if TitlePos <> 0 then
   begin
      DelimeterPos := Pos(';', tmpStr);
      if DelimeterPos = 0 then
         StreamTitle := copy(tmpStr, TitlePos + 13, length(tmpStr) - TitlePos - 13)
      else
         StreamTitle := copy(tmpStr, TitlePos + 13, DelimeterPos - TitlePos - 14);
   end else
   begin   // May be a chained OGG stream
      while StrLen(TagP) > 0 do
      begin
         if pos('TITLE=', upperCase(string(TagP))) <> 0 then
         begin
            inc(TagP, 6);
            StreamTitle := string(TagP);
            break;

            inc(TagP, StrLen(TagP) + 1);
         end;
      end;
   end;

   StrPCopy(TitleP, StreamTitle);
   PostMessage(user, WM_GetMeta, 0, longint(@TitleP));
end;


// This procedure is called when a stream reaches the end.
procedure PlayEndSync(SyncHandle : HSYNC; Channel, data, user : DWORD); stdcall;
begin
   PostMessage(user, WM_GetToEnd, 0, 0);
end;

// This procedure is called when downloading of an internet stream is done.
procedure DownloadSync(SyncHandle : HSYNC; Channel, data, user : DWORD); stdcall;
begin
   PostMessage(user, WM_DownLoaded, 0, 0);
end;

// get the handle of visualizer
function TBASSPlayer.GetVisWindowHandle : HWND;
begin
   result := VisWindowHandle;
end;

procedure TBASSPlayer.Error(msg: string);
var
  s: string;
begin
  s := msg + #13#10 + '(error code: ' + IntToStr(BASS_ErrorGetCode) + ')';
  MessageBox(AppHWND, PChar(s), 'Error', MB_ICONERROR or MB_OK);
end;

procedure TBASSPlayer.ClearEffectHandle;
var
   i : integer;
begin
   fladspHandle := 0;
   if FDX8EffectReady then
   begin
      for i := 0 to (NumEQBands-1) do
          EQHandle[i] := 0;
      EchoHandle := 0;
      ReverbHandle := 0;
   end;
end;

function TBASSPlayer.IsPluginReady : boolean;
begin
   result := OutputReady; // true if Winamp output plug-in emulator is ready
end;

function TBASSPlayer.PlayLength : DWORD;  // get playback length in mili seconds
var
   SongLength : int64;
   FloatLen : FLOAT;
begin
   result := 0;
   case ChannelType of
      Channel_NotOpened : exit;
      Channel_WMA, Channel_Stream : SongLength := BASS_StreamGetLength(Channel);
      Channel_Music  : SongLength := BASS_MusicGetLength(Channel, true);
      Channel_Plugin : begin
                         result := FStreamInfo.Duration;
                         exit;
                       end;
   end;

   if SongLength < 0 then
      SongLength := 0;

   FloatLen := BASS_ChannelBytes2Seconds(Channel, SongLength);
   result := round(1000 * FloatLen);   // sec -> milli sec
end;


function TBASSPlayer.CurrentPosition : DWORD;  // get playback position in mili seconds
var
   SongPos : int64;
   FloatPos : FLOAT;
begin
   result := 0;
   case ChannelType of
      Channel_NotOpened : exit;
      Channel_WMA, Channel_Stream : SongPos := BASS_ChannelGetPosition(Channel);
      Channel_Music  : exit;   // not implemented yet
      Channel_Plugin : begin
                         if FPlayerMode = plmReady then
                            result := 0
                         else if UsesOutputPlugin then
                            result := GetOutputTime(true)  // Output plug-in's time
                         else
                            result := GetOutputTime(false); // Input plug-in's time
                         exit;
                       end;
   end;

   if SongPos < 0 then
      SongPos := 0;

   FloatPos := BASS_ChannelBytes2Seconds(Channel, SongPos);
   result := round(1000 * FloatPos);     // sec -> milli sec
end;


procedure TBASSPlayer.SetPosition(Value : DWORD); // set playback position in mili seconds
var
   SongPos : int64;
   wCycle : integer;
begin
   if ChannelType = Channel_NotOpened then
      exit;

   if Value > PlayLength {= FStreamInfo.Duration} then
      exit;

   if NowStarting then
      exit;

   SongPos := BASS_ChannelSeconds2Bytes(Channel, Value / 1000);
   case ChannelType of
      Channel_NotOpened : exit;
      Channel_Music   : exit;   // not implemented yet
      Channel_WMA, Channel_Stream  : BASS_ChannelSetPosition(Channel, SongPos);
      Channel_Plugin  : begin
                           if (FPlayerMode = plmPlaying) and UsesOutputPlugin then
                              NowStarting := true;
                      //     PostMessage(MessageHandle2, WM_PosChange, 0, 0);
                           SetPosChanged;
                           APlugin.SetOutputTime(Value);
                           wCycle := 0;
                           while FPlayerMode = plmPlaying do
                           begin
                              if NowStarting = false then
                               begin
                                  TimerFFT.Enabled := true;
                                  break;
                               end;
                               WinProcessMessages;
                               sleep(40);
                               inc(wCycle);
                               if wCycle > 50 then   // 2 sec time out
                                  break;
                           end;
                           NowStarting := false;
                        end;
   end;
end;

function TBASSPlayer.GetDownloadProgress : DWORD;  // get downloaded bytes
begin
   if ChannelType = Channel_NotOpened then
      result := 0
   else if (not NetStream) then
      result := 0
   else if ChannelType = Channel_Plugin then  // not available if current stream is decoded by Winamp plug-in
      result := 0
   else
      result := BASS_StreamGetFilePosition(Channel, BASS_FILEPOS_DOWNLOAD);
end;

function TBASSPlayer.IsSeekable : boolean;
begin
   if (ChannelType = Channel_NotOpened) or (ChannelType = Channel_Music) then
      result := false
   else if NetRadio then
      result := false
   else if ChannelType = Channel_Plugin then
      result := (APlugin.is_seekable <> 0)
   else if NetStream then
      result := FDownLoaded
   else
      result := true;
end;


procedure TBASSPlayer.SetVersionStr(Value : string);
begin
   // nothing to do : only to show FVersionStr at form design
end;

procedure TBASSPlayer.SetDLLVersionStr(Value : string);
begin
   // nothing to do : only to show FDLLVersionStr at form design
end;

procedure TBASSPlayer.SetDX8EffectReady(Value : boolean);
begin
   // nothing to do : only to show FDX8EffectReady at form design
end;

procedure TBASSPlayer.SetVolume(Value : DWORD);
var
   tmpValue : DWORD;
begin
   if Value > MaxVolume then
      tmpValue := MaxVolume
   else
      tmpValue := Value;

   FOutputVolume := tmpValue;
   BASS_SetConfig(BASS_CONFIG_GVOL_MUSIC, tmpValue);
   BASS_SetConfig(BASS_CONFIG_GVOL_STREAM, tmpValue);
end;

procedure TBASSPlayer.SetEQGains(Value : TEQGains);
var
   i : integer;
begin
   if not FDX8EffectReady then
      exit;

   for i := 0 to (NumEQBands - 1) do
   begin
      if Value[i] > 15.0 then
         FEQGains[i] := 15.0
      else if Value[i] < -15.0 then
         FEQGains[i] := -15.0
      else
         FEQGains[i] := Value[i];

      if not (Equalizer in FSoundEffects) then
         Continue;

      if EQHandle[i] = 0 then
         Continue;

      BASS_FXGetParameters(EQHandle[i], @EQParam);

   // Avoid redundant operation
      if abs(EQParam.fGain - FEQGains[i]) < 0.001 then   // consider round-off error
         Continue;

      EQParam.fBandwidth := FEQBands.BandPara[i].Bandwidth;
      EQParam.fCenter := FEQBands.BandPara[i].CenterFreq;
      EQParam.fGain:= FEQGains[i];
      BASS_FXSetParameters(EQHandle[i], @EQParam);
   end;
end;

procedure TBASSPlayer.SetEchoLevel(Value : word);
begin
   if Value > MaxDX8Effect then
      exit;

   FEchoLevel := Value;
   if Echo in FSoundEffects then
   begin
       if EchoHandle <> 0 then
       begin
         if BASS_FXGetParameters(EchoHandle, @EchoParam) then
         begin
            EchoParam.fWetDryMix := FEchoLevel * 1.2{30.0};
            EchoParam.fFeedBack := 30.0;
            BASS_FXSetParameters(EchoHandle, @EchoParam)
         end;
       end;
   end;
end;

procedure TBASSPlayer.SetReverbLevel(Value : word);
begin
   if Value > MaxDX8Effect then
      exit;

   FReverbLevel := Value;
   if Reverb in FSoundEffects then
   begin
      if ReverbHandle <> 0 then
       begin
         if BASS_FXGetParameters(ReverbHandle, @ReverbParam) then
         begin
            ReverbParam.fInGain := 0.0;
            ReverbParam.fReverbMix := FReverbLevel * 0.5 - 16.0;
            ReverbParam.fReverbTime := 1000.0;
            ReverbParam.fHighFreqRTRatio := 0.1;
            BASS_FXSetParameters(ReverbHandle, @ReverbParam)
         end;
       end;
   end;
end;

procedure TBASSPlayer.SetEQBands(Value : TEQBands);
var
   i : integer;
begin
   FEQBands.Bands := Value.Bands;
   if FEQBands.Bands > NumEQBands then
      FEQBands.Bands := NumEQBands;

   for i := 0 to (FEQBands.Bands-1) do
   begin
      FEQBands.BandPara[i].CenterFreq := Value.BandPara[i].CenterFreq;
      FEQBands.BandPara[i].Bandwidth := Value.BandPara[i].Bandwidth;
   end;

   if Equalizer in FSoundEffects then
     for i := 0 to (FEQBands.Bands-1) do
     begin
        if EQHandle[i] = 0 then
           EQHandle[i] := BASS_ChannelSetFX(Channel, BASS_FX_PARAMEQ);
        EQParam.fGain := FEQGains[i];
        EQParam.fBandwidth := FEQBands.BandPara[i].Bandwidth;
        EQParam.fCenter := FEQBands.BandPara[i].CenterFreq;
        BASS_FXSetParameters(EQHandle[i], @EQParam);
     end;

   for i := FEQBands.Bands to (NumEQBands-1) do
   begin
      if EQHandle[i] <> 0 then
         if BASS_ChannelRemoveFX(Channel, EQHandle[i]) then
            EQHandle[i] := 0;
   end;
end;

procedure TBASSPlayer.SetSoundEffect(Value : TSoundEffects);
var
   i : integer;
begin
   FSoundEffects := Value;

   if ChannelType = Channel_NotOpened then
      exit;

   if Flanger in Value then
   begin
      if fladspHandle = 0 then
      begin
         FillChar(flabuf, SizeOf(flabuf), 0);
         flapos := 0;
         flas := FLABUFLEN / 2;
         flasinc := 0.002;
         fladspHandle := BASS_ChannelSetDSP(Channel, Flange, 0, 0);
      end;
   end
   else
      if fladspHandle <> 0 then
         if BASS_ChannelRemoveDSP(Channel, fladspHandle) then
            fladspHandle := 0;

   if not FDX8EffectReady then
      exit;

   if Equalizer in Value then
   for i := 0 to (FEQBands.Bands-1) do
   begin
      if EQHandle[i] = 0 then
         EQHandle[i] := BASS_ChannelSetFX(Channel, BASS_FX_PARAMEQ);
      EQParam.fGain := FEQGains[i];
      EQParam.fBandwidth := FEQBands.BandPara[i].Bandwidth;
      EQParam.fCenter := FEQBands.BandPara[i].CenterFreq;
      BASS_FXSetParameters(EQHandle[i], @EQParam);
   end else
   for i := 0 to (NumEQBands-1) do
   begin
      if EQHandle[i] <> 0 then
         if BASS_ChannelRemoveFX(Channel, EQHandle[i]) then
            EQHandle[i] := 0;
   end;

   if Echo in Value then
   begin
       if EchoHandle = 0 then
          EchoHandle := BASS_ChannelSetFX(Channel, BASS_FX_ECHO);
       if EchoHandle <> 0 then
       begin
         if BASS_FXGetParameters(EchoHandle, @EchoParam) then
         begin
            EchoParam.fWetDryMix := FEchoLevel * 1.2{30.0};
            EchoParam.fFeedBack := 30.0;
            BASS_FXSetParameters(EchoHandle, @EchoParam)
         end;
       end;
   end else
      if EchoHandle <> 0 then
         if BASS_ChannelRemoveFX(Channel, EchoHandle) then
            EchoHandle := 0;

   if Reverb in Value then
   begin
       if ReverbHandle = 0 then
          ReverbHandle := BASS_ChannelSetFX(Channel, BASS_FX_REVERB);
       if ReverbHandle <> 0 then
       begin
         if BASS_FXGetParameters(ReverbHandle, @ReverbParam) then
         begin
            ReverbParam.fInGain := 0.0;
            ReverbParam.fReverbMix := FReverbLevel * 0.5 - 16.0;
            ReverbParam.fReverbTime := 1000.0;
            ReverbParam.fHighFreqRTRatio := 0.1;
            BASS_FXSetParameters(ReverbHandle, @ReverbParam)
         end;
       end;
   end else
      if ReverbHandle <> 0 then
         if BASS_ChannelRemoveFX(Channel, ReverbHandle) then
            ReverbHandle := 0;

end;


function  TBASSPlayer.RunVisoutProgram(VismodName : string; VismodNo : word;
                              MainWinHwnd : HWND) : boolean; // * Modified at Ver 1.5
{type
  _PROCESS_INFORMATION = record
     hProcess : HWND;
     hThread : HWND;
     dwProcessId : DWORD;
     dwThreadId : DWORD;
  end; }
var
   TimerStat : boolean;
   CmdLineStr : string;
   ProcId : DWORD;
   VisOutHandle : HWND;
   SyncWord : word;
   StartupInfo: TStartupInfo;
   BeginPath: PChar;
   ProcessInfo : _PROCESS_INFORMATION;
begin
 // Winamp visualization plug-ins are designed to be run in their own user
 // interface thread.
 // I have made a program named "VisOut.exe" to run Winamp visualization plug-ins
 // in a seperate process.
 // The visualization data are fed to "VisOut.exe" via shared memory.
 // note : Some Winamp visualization plug-ins won't be run, if the cooperator
 //        of them is not Winamp.exe. I use fake Winamp window to overcome
 //        this obstacle.

   result := false;

   if VisThread.ShareMemPointer = nil then
   begin
      MessageBox(AppHWND, 'Shared memory for Visout.exe is not obtained', 'Notify',
                                                   MB_OK or MB_ICONWARNING);
      exit;
   end;

   if not FileExists(VismodName) then
   begin
      MessageBox(AppHWND, 'Specified visualization plug-in does not exist', 'Notify',
                                                   MB_OK or MB_ICONWARNING);
      exit;
   end;

   if WaitingVisualizer then   // prohibit duplicate request to launch visualizer
      exit;

   VisOutHandle := FindWindow('TApplication', 'VisOut');
   if VisOutHandle <> 0 then  // VisOut.exe is running ?
   begin
      GetWindowThreadProcessId(VisOutHandle, @ProcId); // get process identifier
      if ProcId = VisProcId then  // if VisOut.exe is being run by this program then quit it
         VisThread.fisSharedMemory1.InformDataReady(QuitProg, 0)
      else begin
      // Multiple instance of VisOut.exe is not allowed because they will use
      // same address space obtained by fisSharedMemory1.
         MessageBox(AppHWND, 'Visout.exe is being run by another program', 'Notify',
                                                   MB_OK or MB_ICONWARNING);
         exit;
      end;
   end;

   WaitingVisualizer := true;
   TimerStat := TimerFFT.Enabled;
   TimerFFT.Enabled := false;

   GetWindowThreadProcessId(AppHWND, @ProcId);
   if FSyncVisWindow then
      SyncWord := 1
   else
      SyncWord := 0;

   CmdLineStr := GetProgDir + 'VisOut.exe'  // Added some parameters at Ver 1.5
                      + ' "' + VismodName
                      + '" '+ intToStr(VismodNo)
                      + ' ' + intToStr(int64(ProcId))
                      + ' ' + intToStr(MainWinHwnd)
                      + ' ' + intToStr(MessageHandle)
                      + ' "' + FStreamInfo.Title
                      + ' " ' + intToStr(FStreamInfo.Channels)
                      + ' ' + intToStr(FStreamInfo.SampleRate)
                      + ' ' + intToStr(FStreamInfo.BitRate)
                      + ' ' + intToStr(FStreamInfo.Duration)
                      + ' ' + intToStr(ord(FPlayerMode))
                      + ' ' + intToStr(SyncWord);

   FillChar (StartupInfo, Sizeof(StartupInfo), 0);
   StartupInfo.cb := Sizeof(StartupInfo);
   StartupInfo.dwFlags := STARTF_USESHOWWINDOW;
   StartupInfo.wShowWindow := SW_NORMAL;
   BeginPath := pchar(GetProgDir);
   Result := CreateProcess(nil, pchar(CmdLineStr),
                           nil, nil, False, CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
                           nil, BeginPath, StartupInfo, ProcessInfo);

   if Result then
   begin
     VisProcId := ProcessInfo.dwProcessId;
     CloseHandle(ProcessInfo.hThread);
     CloseHandle(ProcessInfo.hProcess);
   end else
      MessageBox(AppHWND, 'Failed to run VisOut.exe', 'Error',
                                                   MB_OK or MB_ICONERROR);

   TimerFFT.Enabled := TimerStat;
end;

procedure TBASSPlayer.QuitVisoutProgram;
begin
   VisThread.Suspend;
   GoVisOut := false;

 // Inform "VisOut.exe" that it's time to end
   if FindWindow('TApplication', 'VisOut') <> 0 then
      VisThread.fisSharedMemory1.InformDataReady(QuitProg, 0);

end;

function TBassPlayer.RunDSPPlugin(DSPmodName : string; DSPmodNo : word;
                              MainWinHwnd : HWND) : boolean; // * Added at Ver 1.5
var
   DSPheader : PWinampDSPHeader;
   NumDSPmod : integer;
   Msg : pchar;
begin
   result := false;

   LoadDSPModule(DSPmodName, DSPheader, DSPmod, NumDSPmod, MainWinHwnd);
   if (DSPmodNo < NumDSPmod) then
   begin
      if StartDSPModule(DSPmodNo) = 0 then
      begin
    // DSP plug-ins are activated by direct coupling of input~DSP plug-ins.
    // So, following 'BASS_ChannelSetDSP' is not necessary.
    // If you want to activate DSP plug-ins without the need of input plug-ins,
    // you must complete the WinampDSPProc procedure. ( I have not completed it yet.)
    //     DSPHandle := BASS_ChannelSetDSP(Channel, WinampDSPProc, 0, 0);
         DSPModNum := DSPModNo;
         result := true;   // succeed
      end
      else begin
         UnloadDSPModule;
         Application.MessageBox('Specified plug-in is not operative', 'Error',
                                                          MB_OK + MB_ICONERROR);
         Application.Terminate;
      end;
   end else begin
      if NumDSPmod = 0 then
         Msg := 'Specified plug-in is invalid'
      else
         Msg := 'Invalid module number';
      Application.MessageBox(Msg, 'Error', MB_OK + MB_ICONERROR);
      Application.Terminate;
   end;
end;

procedure TBASSPlayer.QuitDSPPlugin;
begin
   StopDSPModule;
   if DSPHandle <> 0 then
   begin
      BASS_ChannelRemoveDSP(Channel, DSPHandle);
      DSPHandle := 0;
   end;
end;

procedure TBASSPlayer.SetPluginFirst(Value : boolean);
begin
   FPluginFirst := Value;
end;

// This procedure is used to inhibit or to allow unloading of Winamp
// input plug-ins.
//   SetUnloadBut(false) -> inhibit,  SetUnloadBut(true) -> allow
procedure TBASSPlayer.SetUnloadBut(Value : boolean);
begin
   FUnloadButEnabled := value;
   PluginConfigForm.btnUnload.Enabled := value;
end;

procedure TBASSPlayer.SetSyncVisWindow(Value : boolean);
begin
   if Value <> FSyncVisWindow then
   begin
      FSyncVisWindow := Value;
      InformPlayerStatus(stSyncWindows);  // New method
   end;
end;

procedure TBASSPlayer.SetVisScale(Value : word);
begin
   FVisScale := Value;
   VisThread.SetVisScale(FVisScale);
end;


constructor TBASSPlayer.Create(AOwner: TComponent);
var
   i : integer;
   BASSVersion : DWORD;
begin
   inherited Create(AOwner);

   FVersionStr := Ver;
   FBASSReady := false;
   FBASSWMAReady := false;
   ChannelType := Channel_NotOpened;
   FSoundEffects := [];
   FPlayerMode := plmStandby;
   FStreamName :='';
   FStreamInfo.FileName := '';
   FStreamInfo.SampleRate := 0;
   FPluginFirst := false;
   FVisScale := 512;        // Scale factor for visualizer
   FSyncVisWindow := false; // Syncronized action (disabled)

   if not (csDesigning in ComponentState) then
   begin
     AppHWND := Application.Handle;
     MessageHandle := AllocateHWnd(ProcMessage);
     SetMainHandle(MessageHandle);  // set hMainWindow in PluginCtrl unit
     TimerFFT := TTimer.Create(nil);
     TimerFFT.Interval := 25;       // * Changed at Ver 1.51 (33 -> 25)
     TimerFFT.Enabled := false;
     TimerFFT.OnTimer := TimerFFTTimer;
   end;

   BASSDLLLoaded := Load_BASSDLL(GetProgDir + 'bass.dll');
   if not BASSDLLLoaded then
   begin
      if (csDesigning in ComponentState) then
         FDLLVersionStr := 'N.A.(Copy bass.dll in <directory Delphi installed>\Bin to get version)'
      else
         FDLLVersionStr := '';
      exit;
   end;

   BASSVersion := BASS_GetVersion;
   if (BASSVersion <> MAKELONG(2,0)) then
   begin
     if not (csDesigning in ComponentState) then
        Application.MessageBox('BASS version is not 2.0 !', 'Confirm', MB_OK);
     exit;
   end;

   BASS_SetConfig(BASS_CONFIG_MAXVOL, MaxVolume);  // Set maximum volume range
 // setup output - default device, 44100hz, stereo, 16 bits
   if not BASS_Init(1, 44100, 0, AppHWND, nil) then
   begin
      if not (csDesigning in ComponentState) then
         Error('Can''t initialize device');
      exit;
   end else
      FBASSReady := true;

   BassInfoParam.size := SizeOf(BassInfoParam);
   BASS_GetInfo(BassInfoParam);

 // BASS_ChannelSetFX requires DirectX version 8 or higher
   if BassInfoParam.dsver >= 8 then
   begin
      FDX8EffectReady := true;
   end else begin
      if not (csDesigning in ComponentState) then
         MessageBox(AppHWND, NoDX8Msg, 'Warning', MB_ICONWARNING or MB_OK);
      FDX8EffectReady := false;
   end;

   FEQBands.Bands := NumEQBands;
   for i := 0 to (NumEQBands-1) do
   begin
      FEQGains[i] := 0;
      FEQBands.BandPara[i].CenterFreq := EQFreq[i];
      FEQBands.BandPara[i].Bandwidth := EQBandWidth[i];
   end;
   FEchoLevel := MaxDX8Effect div 2{16};
   FReverbLevel := MaxDX8Effect div 2{16};

   FOutputVolume := MaxVolume;
   BASS_SetConfig(BASS_CONFIG_GVOL_MUSIC, FOutputVolume);
   BASS_SetConfig(BASS_CONFIG_GVOL_STREAM, FOutputVolume);

   FDLLVersionStr := GetDLLVersionStr;

   if not (csDesigning in ComponentState) then
   begin
      if FBASSReady then
      begin
         if Load_BASSWMADLL(GetProgDir + 'basswma.dll') then
            FBASSWMAReady := true;
      end;

      InitPluginCtrl;
      MessageHandle2 := GetMsgHandler;  // = message handler of PluginCtrl unit
      MPEG := TMPEGaudio.Create;
      Vorbis := TOggVorbis.Create;
      WMA := TWMAfile.Create;
      WAV := TWAVFile.Create;

  // Is it right to create a form component in this way ?
  // Anyway I do not know the better way and this method works.
      PluginConfigForm := TPluginConfigForm.Create(Self);
      FUnloadButEnabled := PluginConfigForm.btnUnload.Enabled;

      VisThread := TVisThread.Create(Self);
      VisThread.SetVisScale(FVisScale);
   end;
end;


destructor  TBASSPlayer.Destroy;
begin
   if ChannelType = Channel_Plugin  then
   begin
      if Channel <> 0 then
         APlugin.Stop
   end else if (FPlayerMode = plmPlaying) then
      Stop;

   repeat
      WinProcessMessages;
   until BASS_ChannelIsActive(Channel) <> BASS_ACTIVE_PLAYING;

   if not (csDesigning in ComponentState) then
   begin
     TimerFFT.Free;
     MPEG.Free;
     Vorbis.Free;
     WMA.Free;
     WAV.Free;
     QuitPluginCtrl;
     if MessageHandle <> 0 then
        DeallocateHWnd(MessageHandle);

     VisThread.Free;
     PluginConfigForm.Free;
   end;

   if BASSDLLLoaded then
   begin
      BASS_Free;
      Unload_BASSDLL;
   end;
   if FBASSWMAReady then
      Unload_BASSWMADLL;

   inherited Destroy;
end;

// Get the version information as string
function TBASSPlayer.GetDLLVersionStr : string;
var
   VersionNum : DWORD;
begin
   if not FBASSReady then
   begin
      result := '';
      exit;
   end;

   VersionNum := BASS_GetVersion;
   result := intToStr(LoWord(VersionNum)) + '.' + intToStr(HiWord(VersionNum));
end;

function TBASSPlayer.GetNativeFileExts : string;
begin
   if FBASSReady and FBASSWMAReady then
      result := '*.wav;*.mp1;*.mp2;*.mp3;*.ogg;*.wma;'
   else if FBASSReady then
      result := '*.wav;*.mp1;*.mp2;*.mp3;*.ogg;'
   else
      result := '';
end;

function TBASSPlayer.GetPluginFileExts : string;
begin
   result := GetPlayableFiles;
end;

function TBASSPlayer.GetStreamInfo2(StreamName : string;
                       var StreamInfo : TStreamInfo;
                       var SupportedBy : TSupportedBy;
                       var PluginNum : integer) : boolean;
var
   i : integer;
   s, ExtCode : string;
   APlugin : TPlugin;
   APluginInfo : TPluginInfo;
   _file : array[0..255] of char;
   _title : array[0..255] of char;
   _length_in_ms : integer;
begin
   result := false;

   with StreamInfo do
   begin
      FileName := '';
      FileSize := 0;
      SampleRate := 0;
      BitRate := 0;
      BitsPerSample := 16;  // default value except WAV file
      Duration := 0;
      Title := '';
      Artist := '';
      Album := '';
      Year := '';
      Genre := '';
      GenreID := 0;
      Track := 0;
      Comment := '';
      Channels := 0;
   end;

   StreamInfo.FileName := StreamName;
   SupportedBy := None;
   PluginNum := -1;

   ExtCode := UpperCase(ExtractFileExt(StreamName));
   if (ExtCode = '') or (length(ExtCode) <> 4) then
      exit;

 // Check if native file types
   if ExtCode = '.WAV' then
   begin
      WAV.ReadFromFile(StreamName);
      if WAV.Valid then
      begin
         with StreamInfo do
         begin
            FileSize := WAV.FileSize;
            SampleRate := WAV.SampleRate;
            BitsPerSample := WAV.BitsPerSample;
            BitRate := (SampleRate * BitsPerSample) div 1000;  // Why not 1024 ?
            Duration := round(WAV.Duration * 1000);
            Channels := WAV.ChannelModeID;
         end;
         if FBASSReady then
            SupportedBy := BASSNative;
         result := true;
      end;
   end else if (ExtCode = '.MP1') or (ExtCode = '.MP2') or (ExtCode = '.MP3') then
   begin
      MPEG.ReadFromFile(StreamName);
      if MPEG.Valid then
      begin
         with StreamInfo do
         begin
            FileSize := MPEG.FileLength;
            SampleRate := MPEG.SampleRate;
            BitRate := MPEG.BitRate;
            Duration := round(MPEG.Duration * 1000);

            if MPEG.ID3v1.Exists then
            begin
               Title := MPEG.ID3v1.Title;
               Artist := MPEG.ID3v1.Artist;
               Album := MPEG.ID3v1.Album;
               Year := MPEG.ID3v1.Year;
               Genre := MPEG.ID3v1.Genre;
               GenreID := MPEG.ID3v1.GenreID;
               Track := MPEG.ID3v1.Track;
               Comment := MPEG.ID3v1.Comment;
            end else if MPEG.ID3v2.Exists then
            begin
               Title := MPEG.ID3v2.Title;
               Artist := MPEG.ID3v2.Artist;
               Album := MPEG.ID3v2.Album;
               Year := MPEG.ID3v2.Year;
               Genre := MPEG.ID3v2.Genre;
               Track := MPEG.ID3v2.Track;
               Comment := MPEG.ID3v2.Comment;
            end;

            if MPEG.ChannelMode = 'Mono' {MPEG_CM_MONO} then
               Channels := 1
            else
               Channels := 2;
         end;
         if FBASSReady then
            SupportedBy := BASSNative;
         result := true;
      end;
   end else if ExtCode = '.OGG' then
   begin
      Vorbis.ReadFromFile(StreamName);
      if Vorbis.Valid then
      begin
         with StreamInfo do
         begin
            FileSize := Vorbis.FileSize;
            SampleRate := Vorbis.SampleRate;
            BitRate := Vorbis.BitRate;
            Duration := round(Vorbis.Duration * 1000);
            Title := Vorbis.Title;
            Artist := Vorbis.Artist;
            Album := Vorbis.Album;
            Year := Vorbis.Date;
            Genre := Vorbis.Genre;
            Channels := Vorbis.ChannelModeID;
            Track := Vorbis.Track;
            Comment := Vorbis.Comment;
         end;
         if FBASSReady then
            SupportedBy := BASSNative;
         result := true;
      end;
   end else if ExtCode = '.WMA' then
   begin
      WMA.ReadFromFile(StreamName);
      if WMA.Valid then
      begin
         with StreamInfo do
         begin
            FileSize := WMA.FileSize;
            SampleRate := WMA.SampleRate;
            BitRate := WMA.BitRate;
            Duration := round(WMA.Duration * 1000);
            Title := WMA.Title;
            Artist := WMA.Artist;
            Album := WMA.Album;
            Year := WMA.Year;
            Genre := WMA.Genre;
            Channels := WMA.ChannelModeID;
            Track := WMA.Track;
            Comment := WMA.Comment;
         end;
         if FBASSWMAReady then
            SupportedBy := BASSNative;
         result := true;
      end;
   end;

 // Check if it is the file type that can be played by Winamp input plug-in
   for i := 0 to MaxPluginNum - 1 do
   begin
      APlugin := GetPlugin(i);
      if APlugin <> nil then
      begin
         APluginInfo := GetPluginInfo(i);
         s := uppercase(APluginInfo.FileExtensions);
         if pos(copy(ExtCode, 2, 3), s) <> 0 then
         begin
            PluginNum := i;
            StrPCopy(_file, StreamName);
            APlugin.GetFileInfo(_file, _title, _length_in_ms);
            if (_length_in_ms > 0) then
            begin
               with StreamInfo do
               begin
                  if SupportedBy = BASSNative then
                     SupportedBy := Both
                  else begin
                     if FBASSReady then
                        SupportedBy := WinampPlugin;
                     Title := string( _title);
                     Duration := _length_in_ms;
                     result := true;
                  end;
                  break;
               end;
            end;
         end;
      end;
   end;

   if (result = true) and (StreamInfo.Title = '') then
      StreamInfo.Title := ExtractFileName(StreamName);
end;

function TBASSPlayer.GetStreamInfo(StreamName : string;
                       var StreamInfo : TStreamInfo;
                       var SupportedBy : TSupportedBy) : boolean;
var
   dumNum : integer;
begin
   result := GetStreamInfo2(StreamName, StreamInfo, SupportedBy, dumNum);
end;


procedure TBASSPlayer.ProcMessage(var Msg: TMessage);
var
   PChannelInfo : ^TChannelInfo;
   PTitle : PChar;
   ExtCode : string;
   TagP : pchar;
   TagVer : word;
   MP3Tag : MP3TagRec;
   PlaybackSec : float;

   _title : array[0..255] of char;
   _length_in_ms : integer;
begin
   if Msg.Msg = VisThread.DataReadyMsg then   // message from visualizer ?
   begin
      if Msg.wParam = VisModuleLoaded then      // visualizer loaded ?
      begin
         VisThread.SetDelayMS(Msg.lParam);
      end else if Msg.wParam = StartVisOut then // visualizer started ?
      begin
         if Msg.lParam <> 0 then
         begin
            VisWindowHandle := Msg.lParam;
            VisThread.fisSharedMemory1.TargetApplication := 'VisOut';
            GoVisOut := true;
            if (FPlayerMode = plmPlaying) and VisThread.Suspended then
               VisThread.Resume; 
         end;
         WaitingVisualizer := false;
      end else
      if Msg.wParam = EndVisOut then       // visualizer ended ?
      begin
         VisWindowHandle := 0;
         GoVisOut := false;
         if not VisThread.Suspended then
            VisThread.Suspend; 
      end else
      if Msg.wParam = QueryStatus then      // query from visualizer ?
      begin
         if Msg.lParam = stPlayerMode then
            InformPlayerStatus(stPlayerMode)
         else if Msg.lParam = stStreamInfo then
            InformPlayerStatus(stStreamInfo)
         else if Msg.lParam = stSyncWindows then
            InformPlayerStatus(stSyncWindows);
      end;

      exit;
   end;

   case Msg.Msg of
      WM_NewFFTData    : if Assigned(FOnNewFFTData) then
                            FOnNewFFTData(Self, BandOut);
   // We do not know the Channels & SampleRate of opened stream file before Winamp
   // input plug-in informs to output plug-in, if the file type is the one which is only
   // supported by Winamp input plug-in.
   // WM_StartPlay message is used to receive the Channels & SampleRate of opened
   // stream file at starting playback from output plug-in emulator.
   // We still cannot know the Channels & SampleRate of opened stream file if its related
   // Winamp input plug-in is the type which directly sound out (= not thru output plug-in emulator).
      WM_StartPlay     : begin             // Message from output plug-in emulator
                            NowStarting := false;
                            UsesOutputPlugin := true;
                            SetVolume(FOutputVolume);
                            PChannelInfo := pointer(Msg.lParam);
                            FStreamInfo.Channels := PChannelInfo^.Channels;
                            FStreamInfo.SampleRate := PChannelInfo^.SampleRate;
                            FStreamInfo.BitsPerSample := PChannelInfo^.BitsPerSample;
                            if FStreamInfo.Duration = 0 then  // for internet stream
                            begin
                           //    FStreamInfo.Duration := APlugin.GetLength;
                               APlugin.GetFileInfo(nil, _title, _length_in_ms);
                               if (_length_in_ms > 0) then
                               begin
                                  with FStreamInfo do
                                  begin
                                    Title := string( _title);
                                    Duration := _length_in_ms;
                                  end;
                               end;
                            end;
                            InformPlayerStatus(stStreamInfo);
                            if Assigned(OnPluginStartPlay) then
                               FOnPluginStartPlay(Self, pointer(Msg.lParam));
                            VisThread.SetChannelInfo(Channel,
                                                     FStreamInfo.SampleRate,
                                                     FStreamInfo.Channels); 
                            if Assigned(FOnNewFFTData) then
                               TimerFFT.Enabled := true;
                            SetSoundEffect(FSoundEffects);
                         end;
      WM_GetChannelInfo : begin    // The basic characteristics of a stream to be played
                                   // is received from Winamp input plug-in.
                            PChannelInfo := pointer(Msg.lParam);
                            FStreamInfo.Channels := PChannelInfo^.Channels;
                            FStreamInfo.SampleRate := PChannelInfo^.SampleRate;
                            FStreamInfo.BitRate := PChannelInfo^.BitRate;
                            if Assigned(OnGetChannelInfo) then
                               FOnGetChannelInfo(Self, pointer(Msg.lParam));
                         end;
      WM_GetMeta       : begin     // BASS received Metadata in a Shoutcast stream
                            PTitle := pointer(Msg.lParam);
                            FStreamInfo.Title := string(PTitle);
                            InformPlayerStatus(stStreamInfo);
                            if Assigned(OnGetMeta) then
                               FOnGetMeta(Self, FStreamInfo.Title);
                         end;
      WM_DownLoaded    : begin     // BASS finished download of a internet stream
                            FDownloaded := true;
                            ExtCode := UpperCase(ExtractFileExt(FStreamName));
                            if (ExtCode = '.MP3') or (ExtCode = '.MP2') or (ExtCode = '.MP1') then
                            begin
                               TagP := BASS_StreamGetTags(Channel, BASS_TAG_ID3);
                               if TagP = nil then
                               begin
                                  TagP := BASS_StreamGetTags(Channel, BASS_TAG_ID3V2);
                                  if TagP = nil then
                                     TagVer := 0
                                  else
                                     TagVer := 2;
                               end else
                                  TagVer := 1;
                            end;

                            FStreamInfo.FileSize := BASS_StreamGetFilePosition(Channel, BASS_FILEPOS_END);

                            if MPEG.ReadFromTagStream(TagP, FStreamInfo.FileSize, TagVer, MP3Tag) then
                               with FStreamInfo do
                               begin
                                  Title := MP3Tag.Title;
                                  Artist := MP3Tag.Artist;
                                  Album := MP3Tag.Album;
                                  Year := MP3Tag.Year;
                                  Genre := MP3Tag.Genre;
                                  Track := MP3Tag.Track;
                                  Comment := MP3Tag.Comment;
                               end;

                         // Now we can get the exact value of playback length of an internet stream after finishing download.
                            PlaybackSec := BASS_ChannelBytes2Seconds(Channel, BASS_StreamGetLength(Channel));
                            FStreamInfo.Duration := round(1000 * PlaybackSec);  // in mili seconds
                            if Assigned(OnDownloaded) then
                               FOnDownloaded(Self, IntToStr(FStreamInfo.Duration));
                         end;
      WM_BASS_StreamCreate : Channel := Msg.WParam;  // Message from output plug-in emulator
     { WM_BASS_StreamFree   : begin                   // Message from output plug-in emulator
                                Channel := 0;
                                ClearEffectHandle;
                             end; }
      WM_WA_MPEG_EOF   : begin           // Winamp input plug-in reached the end of playback
                         //   SetPlayEnded;
                            NowStarting := false;
                            if UsesOutputPlugin then
                               while BASS_ChannelIsActive(Channel) = BASS_ACTIVE_PLAYING do
                               begin
                                  WinProcessMessages;
                                  sleep(50);
                               end;

                            Stop;
                            if Assigned(FOnPlayEnd) then
                               FOnPlayEnd(Self);
                         end;
      WM_GetToEnd      : begin                       // BASS reached the end of playback
                            FPlayerMode := plmStopped;
                            InformPlayerStatus(stPlayerMode);
                            if not VisThread.Suspended then
                               VisThread.Suspend; 
                            if Assigned(FOnPlayEnd) then
                               FOnPlayEnd(Self);
                         end;
      WM_PluginFirst_Changed : if Msg.lParam = 0 then  // Message from PluginConfigForm
                                  FPluginFirst := false
                               else
                                  FPluginFirst := true;

      WM_QueryEndSession : Msg.Result := 1;    // Allow system termination
   end;
end;


function TBASSPlayer.Open(StreamName : string) : boolean;
var
   FromNet : boolean;
   StreamInfo : TStreamInfo;
   SupportedBy : TSupportedBy;
   PluginNum : integer;
   NameHeader1, NameHeader2 : string;
   tmpName : string;
   TagP : pchar;
   PlaybackSec : float;
   FileLen : integer;
   ExtCode, tmpStr : string;
   TitlePos, DelimeterPos : integer;

 // get the file name from the name of an internet stream
 // (ex: http://.../.../file_name.mp3 -> file_name)
  function ExtractFileName2(NetStream : string) : string;
  var
    pos, i : integer;
  begin
     pos := 0;
     for i := (length(NetStream) - 6)  downto 6 do
        if NetStream[i] = '/' then
        begin
           pos := i;
           break;
        end;

     if pos > 0 then
        result := copy(NetStream, pos + 1, length(NetStream) - pos - 4)
     else
        result := '';
  end;

begin
   result := false;

   if not FBASSReady then
   begin
      Application.MessageBox('Player is not ready !', 'Confirm', MB_OK);
      exit;
   end;

   FromNet := false;
   NameHeader1 := copy(StreamName, 1, 7);
   NameHeader2 := copy(StreamName, 1, 6);
   if (NameHeader1 = 'http://') or (NameHeader2 = 'ftp://') or
      (NameHeader2 = 'mms://') then
   begin
      FromNet := true;
      StreamInfo.FileName := StreamName;
      StreamInfo.BitRate := 0;
   end
   else if not GetStreamInfo2(StreamName, StreamInfo, SupportedBy, PluginNum) then
   begin
      Application.MessageBox(pchar('Invalid stream file (' + StreamName + ') !'),
                                                     'Confirm', MB_OK);
      exit;
   end;

   // release previously assigned channel
   case ChannelType of
      Channel_Music : BASS_MusicFree(Channel);
      Channel_WMA, Channel_Stream : BASS_StreamFree(Channel);
      Channel_Plugin : if Channel <> 0 then
                          APlugin.Stop;
   end;

   ChannelType := Channel_NotOpened;
   FDecodingByPlugin := false;
   FDownLoaded := false;
   NetStream := false;
   NetRadio  := false;
   ICYTag    := nil;
   HTTPTag   := nil;

   if FromNet then   // It's an internet stream.
   begin
      ExtCode := UpperCase(ExtractFileExt(StreamName));

  // ASF files and WMA files are streamed with Winamp input plug-in for the
  // file types.
  // note) WMA files on the internet can be streamed without Winamp plug-in, but
  //       I have found that some of BASS functions do not work properly.
  //       So I have decided to use Winamp plug-in.

      if (ExtCode = '.ASF') or (ExtCode = '.WMA') then
      begin
         if ExtCode = '.ASF' then
            PluginNum := GetPluginNumber('ASF')
         else
            PluginNum := GetPluginNumber('WMA');
         if PluginNum >= 0 then  // = A plug-in which can play ASF (or WMA) files is loaded
         begin
            ChannelType := Channel_Plugin;
            FPluginNum := PluginNum;
            APlugin := GetPlugin(PluginNum);
            FDecodingByPlugin := true;
            SelectInputPlugin(FPluginNum);
            StreamInfo.SampleRate := 0;
            StreamInfo.Duration := 0;
         end else
         begin
            Application.MessageBox('Winamp input plug-in for ASF (or WMA) file is not loaded.',
                                                       'Confirm', MB_OK);
            exit;
         end;
      end else
      begin
         Channel := BASS_StreamCreateURL(pchar(StreamName), 0, BASS_STREAM_META, nil, 0);
         if Channel <> 0 then
            ChannelType := Channel_Stream
         else if FBASSWMAReady then
         begin
         // Change the URL name "http://..." to "mms://..." {and add add a #0 at the end of the URL,}
         // for that of WMA stream.
         // I do not know the reason for this operation. Anyway BASS_WMA_StreamCreateFile may
         // not work without this operation.
            tmpName := StreamName {+ chr(0)};
            if NameHeader1 = 'http://' then
               tmpName := 'mms://' + copy(tmpName, 8, length(tmpName) - 7);
            Channel := BASS_WMA_StreamCreateFile(FALSE, PChar(tmpName), 0, 0, BASS_SYNC_META);
            if Channel <> 0 then
               ChannelType := Channel_WMA;
         end;
      end;

      if (ChannelType <> Channel_Plugin) and (Channel = 0) then  // not a playable URL stream
      begin
         Application.MessageBox(pchar('Can''t play the specified URL stream : ' + StreamName),
                                  'Confirm', MB_OK);
         Exit;
      end;

      if (ExtCode = '.MP3') or (ExtCode = '.MP2') or (ExtCode = '.MP1')
         or (ExtCode = '.OGG') or (ExtCode = '.WAV') or (ExtCode = '.WMA')
           or (ExtCode = '.ASF') then
      begin
         NetStream := true;

         if (ExtCode = '.OGG') then
         begin
            TagP := BASS_StreamGetTags(Channel, BASS_TAG_OGG);
            while StrLen(TagP) > 0 do
            begin
               tmpStr := string(TagP);
               if pos('TITLE=', tmpStr) <> 0 then
               begin
                  StreamInfo.Title := copy(tmpStr, 7, length(tmpStr) - 6);
                  break;
               end;
               inc(TagP, StrLen(TagP) + 1);
            end;

     { // <-
         end else if (ExtCode = '.WMA') then
         begin
            TagP := BASS_StreamGetTags(Channel, BASS_TAG_ID3V2);
            ICYTag := TagP;
            while StrLen(TagP) > 0 do
            begin
               tmpStr := string(TagP);
               if pos('FileSize : ', tmpStr) <> 0 then
                  StreamInfo.FileSize := strToInt(copy(tmpStr, 12, length(tmpStr) - 11));
               if pos('Bitrate : ', tmpStr) <> 0 then
               begin
                  SepPos := pos(':', tmpStr);
                  StreamInfo.Bitrate := strToInt(copy(tmpStr, SepPos + 2, length(tmpStr) - (SepPos + 1))) div 1000;
               end;
               if pos('Title : ', tmpStr) <> 0 then
                  StreamInfo.Title := copy(tmpStr, 9, length(tmpStr) - 8);

               inc(TagP, StrLen(TagP) + 1);
            end;  // <-   }


      // We can get the TITLE & Playback length of stream files on the internet
      // at just before the starting playback from Winamp input plug-in if decoded
      // by Winamp input plug-in.
         end else if (ChannelType <> Channel_Plugin) then
            StreamInfo.Title := ExtractFileName2(StreamName);

         if ChannelType <> Channel_Plugin then
         begin
            if (NameHeader1 = 'http://') or (NameHeader2 = 'mms://') then
            begin
               TagP := BASS_StreamGetTags(Channel, BASS_TAG_HTTP);
               HTTPTag := TagP;
            end;

         // BASS_StreamGetLength may return a very rough estimation until the whole file
         // has been downloaded.
         // So it is needed to re-estimate after finishing download.
            PlaybackSec := BASS_ChannelBytes2Seconds(Channel, BASS_StreamGetLength(Channel));
            StreamInfo.Duration := round(1000 * PlaybackSec);  // in mili seconds

         // note) BASS_StreamGetFilePosition does not work for WMA files on the internet.
         //       Does the BASSWMA always download and play WMA files on the internet in
         //       smaller chunks ? (i.e., operates only in "buffered" mode ?)
            FileLen := BASS_StreamGetFilePosition(Channel, BASS_FILEPOS_END); // file length
            if FileLen <> -1 then
            begin
               StreamInfo.FileSize := FileLen;
               StreamInfo.BitRate := round(FileLen / (125 * PlaybackSec));   // bitrate (Kbps)
            end;

            BASS_ChannelSetSync(Channel, BASS_SYNC_DOWNLOAD, 0, @DownloadSync, MessageHandle);
         end;
      end else   // = Shout/Icecast server
      begin
         NetRadio := true;
         TagP := BASS_StreamGetTags(Channel, BASS_TAG_ICY);
         ICYTag := TagP;
         while StrLen(TagP) > 0 do
         begin
           if pos('icy-br:', string(TagP)) <> 0 then
           begin
              inc(TagP, 7);
              StreamInfo.BitRate := strToInt(string(TagP));
              break;
           end;
           inc(TagP, StrLen(TagP) + 1);
         end;

         TagP := BASS_StreamGetTags(Channel, BASS_TAG_META);
         tmpStr := string(TagP);
         TitlePos := Pos('StreamTitle=', tmpStr);
         if TitlePos <> 0 then
         begin
            DelimeterPos := Pos(';', tmpStr);
            if DelimeterPos = 0 then
               StreamInfo.Title := copy(tmpStr, TitlePos + 13, length(tmpStr) - TitlePos - 13)
            else
               StreamInfo.Title := copy(tmpStr, TitlePos + 13, DelimeterPos - TitlePos - 14);
         end else
            StreamInfo.Title := '';

         BASS_ChannelSetSync(Channel, BASS_SYNC_META, 0, @MetaSync, MessageHandle);
      end;

      if (ChannelType <> Channel_Plugin) then
      begin
         BASS_ChannelGetInfo(Channel, BassChannelInfo);
         StreamInfo.SampleRate := BassChannelInfo.freq;
         StreamInfo.Channels := BassChannelInfo.chans;
         VisThread.SetChannelInfo(Channel, StreamInfo.SampleRate, StreamInfo.Channels);
      end;
   end    // end of for 'internet stream'

 // for local stream files
   else if (SupportedBy = WinampPlugin) or
      ((SupportedBy = Both) and FPluginFirst) then
   begin
      ChannelType := Channel_Plugin;
      FPluginNum := PluginNum;
      APlugin := GetPlugin(PluginNum);
  //    PostMessage(MessageHandle2, WM_StreamOpen, MessageHandle, FPluginNum);
      SelectInputPlugin(FPluginNum);
      FDecodingByPlugin := true;
   end else
   begin
      Channel := BASS_StreamCreateFile(FALSE, PChar(StreamName), 0, 0, 0{BASS_SAMPLE_FX});
      if (Channel <> 0) then
         ChannelType := Channel_Stream
      else
      begin
         if FBASSWMAReady then
            Channel := BASS_WMA_StreamCreateFile(FALSE, PChar(StreamName), 0, 0, 0{BASS_SAMPLE_FX}) // it's a WMA?
         else
            Channel := 0;
         if (Channel <> 0) then
            ChannelType := Channel_WMA
         else begin
            Channel := BASS_MusicLoad(FALSE, PChar(StreamName), 0, 0, BASS_MUSIC_CALCLEN, 0);  // * changed at Ver 1.3 (added flag BASS_MUSIC_CALCLEN)
            if (Channel <> 0) then
               ChannelType := Channel_Music
            else
               ChannelType := Channel_NotOpened;
         end;
      end;

      if (Channel = 0) then  // not a playable file
      begin
         Error('Can''t play the file');
         Exit;
      end;

   //   chattr := BASS_ChannelGetFlags(Channel) and ({BASS_SAMPLE_MONO or} BASS_SAMPLE_8BITS);
      BASS_ChannelGetInfo(Channel, BassChannelInfo);
      if (BassChannelInfo.flags and BASS_SAMPLE_8BITS) = BASS_SAMPLE_8BITS then
      begin
      // does not support 16-bit sources.
         BASS_MusicFree(Channel);
         BASS_StreamFree(Channel);
         Application.MessageBox('Sound card does not support 16-bit sources.',
                                  'Confirm', MB_OK);
         Exit;
      end;
      VisThread.SetChannelInfo(Channel, StreamInfo.SampleRate, StreamInfo.Channels);
   end;   // end of for 'local stream'

  // Reset previous handle values
   ClearEffectHandle;

  // Apply equalyzation and sound effects
   if ChannelType <> Channel_Plugin then
      SetSoundEffect(FSoundEffects);

   FStreamName := StreamName;
   FStreamInfo := StreamInfo;
   InformPlayerStatus(stStreamInfo);
   FPlayerMode := plmReady;
   InformPlayerStatus(stPlayerMode);
   result := true;
end;


procedure TBASSPlayer.Play;
var
   StartOK : boolean;
   wCycle : integer;
begin
   if NowStarting then   // avoid duplicate Play operation
      exit
   else if ChannelType = Channel_NotOpened then
      exit
   else if FPlayerMode = plmPlaying then
      exit
   else if FPlayerMode = plmPaused then
   begin
      ResumePlay;
      exit;
   end else if FPlayerMode = plmStopped then
   begin
      if ChannelType <> Channel_Plugin then
         SetPosition(0);   // Set position to starting point
      Restart;
      exit;
   end;

   StartOK := false;

 // play an opened channel
   case ChannelType of
      Channel_Music : StartOK := BASS_MusicPlay(Channel);
      Channel_WMA, Channel_Stream : StartOK := BASS_StreamPlay(Channel, false, 0);
      Channel_Plugin :  begin
                            NowStarting := true;
                            UsesOutputPlugin := false;
                            Channel := 1;  // for Winamp input plug-in which does not need output plug-in
                            SetSongDuration(FStreamInfo.Duration);
                            APlugin.Play(pchar(FStreamName));
                         //   SetVolume(FPluginVolume);
                            wCycle := 0;
                            while wCycle < 60 do
                            begin
                               if GetOutputTime(false) > 200 then
                               begin
                                  StartOK := true;
                                  break;
                               end;
                               WinProcessMessages;
                               inc(wCycle);
                               sleep(50);
                            end;
                            if not StartOK then
                               MessageBox(AppHWND,
                                     PChar('Can''t play the specified stream : ' + FStreamName),
                                     'Error', MB_ICONERROR or MB_OK);
                            NowStarting := false;
                        end;
   end;

   if StartOK then
   begin
      if ChannelType <> Channel_Plugin then
         BASS_ChannelSetSync(Channel, BASS_SYNC_END, 0, @PlayEndSync, MessageHandle);
      FPlayerMode := plmPlaying;
      InformPlayerStatus(stPlayerMode);
      if GoVisOut and VisThread.Suspended then
         VisThread.Resume; 
      if (ChannelType <> Channel_Plugin) then
         if Assigned(FOnNewFFTData) then
            TimerFFT.Enabled := true;
   end;
end;

procedure TBASSPlayer.PausePlay;
begin
   if FPlayerMode <> plmPlaying then
      exit;

   if ChannelType = Channel_Plugin then
      APlugin.Pause
   else
      BASS_ChannelPause(Channel);

   FPlayerMode := plmPaused;
   InformPlayerStatus(stPlayerMode);
   if not VisThread.Suspended then
      VisThread.Suspend; 
end;

procedure TBASSPlayer.ResumePlay;
begin
   if FPlayerMode <> plmPaused then
      exit;

   if ChannelType = Channel_Plugin then
      APlugin.UnPause
   else
      BASS_ChannelResume(Channel);

   FPlayerMode := plmPlaying;
   InformPlayerStatus(stPlayerMode);
   if GoVisOut and VisThread.Suspended then
      VisThread.Resume; 
   if Assigned(FOnNewFFTData) then
      if ChannelType <> Channel_Plugin then
         TimerFFT.Enabled := true
      else if UsesOutputPlugin then
         TimerFFT.Enabled := true;
end;

procedure TBASSPlayer.Pause(pAction : boolean);
begin
   if pAction then
      PausePlay
   else
      ResumePlay;
end;

procedure TBASSPlayer.Stop;
begin
   if FPlayerMode <> plmPlaying then
      exit;

   if ChannelType = Channel_Plugin then
   begin
      if Channel <> 0 then
      begin
         APlugin.Stop;
         Channel := 0;
         ClearEffectHandle;
      end
   end else
      BASS_ChannelStop(Channel);

   FPlayerMode := plmStopped;
   InformPlayerStatus(stPlayerMode);
   if not VisThread.Suspended then
      VisThread.Suspend;
end;

procedure TBASSPlayer.Restart;
var
   StartOK : boolean;
   wCycle : integer;
begin
   if FPlayerMode <> plmStopped then
      exit;
   if NowStarting then
      exit;
   StartOK := false;

  // play a opened channel
   case ChannelType of
       Channel_Music : StartOK := BASS_MusicPlay(Channel);
       Channel_WMA, Channel_Stream : StartOK := BASS_StreamPlay(Channel, false, 0);
       Channel_Plugin : begin
                           NowStarting := true;
                           UsesOutputPlugin := false;
                           Channel := 1;  // for Winamp input plug-in which does not need output plug-in
                           APlugin.Play(pchar(FStreamName));
                       //    SetVolume(FPluginVolume);
                           wCycle := 0;
                           while wCycle < 60 do
                           begin
                               if GetOutputTime(false) > 200 then
                               begin
                                  StartOK := true;
                                  break;
                               end;
                               WinProcessMessages;
                               inc(wCycle);
                               sleep(50);
                           end;
                           NowStarting := false;
                        end;
   end;

   if StartOK then
   begin
      FPlayerMode := plmPlaying;
      InformPlayerStatus(stPlayerMode);
      if GoVisOut and VisThread.Suspended then
         VisThread.Resume; 
      if (ChannelType <> Channel_Plugin) then
         if Assigned(FOnNewFFTData) then
            TimerFFT.Enabled := true;
   end;
end;


procedure TBASSPlayer.Close;
begin
   case ChannelType of
      Channel_Music : BASS_MusicFree(Channel);
      Channel_Stream, Channel_WMA : BASS_StreamFree(Channel);
      Channel_Plugin : begin
                          if Channel <> 0 then
                          begin
                             APlugin.Stop;
                             Channel := 0;
                             ClearEffectHandle;
                          end;
                          FPluginNum := -1;
                          APlugin := nil;
                          PostMessage(MessageHandle2, WM_StreamClose, 0, 0);
                       end;
   end;

   ChannelType := Channel_NotOpened;
   FPlayerMode := plmStandby;
   InformPlayerStatus(stPlayerMode);
   if not VisThread.Suspended then
      VisThread.Suspend; 
   FStreamName :='';
   FStreamInfo.FileName := '';
   FStreamInfo.SampleRate := 0;
end;


function TBASSPlayer.GetChannelFFTData(PFFTData : pointer; FFTFlag : DWORD) : boolean;
begin
   result := false;

   if BASS_ChannelIsActive(Channel) <> BASS_ACTIVE_PLAYING then
      exit;

   if (FFTFlag = BASS_DATA_FFT512) or
      (FFTFlag = BASS_DATA_FFT512 + BASS_DATA_FFT_INDIVIDUAL) or
      (FFTFlag = BASS_DATA_FFT1024) or
      (FFTFlag = BASS_DATA_FFT1024 + BASS_DATA_FFT_INDIVIDUAL) or
      (FFTFlag = BASS_DATA_FFT2048) or
      (FFTFlag = BASS_DATA_FFT2048 + BASS_DATA_FFT_INDIVIDUAL) then
      if BASS_ChannelGetData(Channel, PFFTData, FFTFlag) <> $ffffffff then
         result := true;
end;


// If you have any idea on spectrum visualization for better result,
// please let me know.
procedure TBASSPlayer.TimerFFTTimer(Sender: TObject);
const
 // Classify FFTData[x] by the number of required bands for spectrum visualization.
 // FreqCoef defines the last index number of FFTData[x] for each band.
   FreqCoef : array[0..NumFFTBands - 1] of word =
               (  1,  2,    3,   6,       // narrow interval for low freq. range
                 12,  18,  24,  30,  36,  // normal interval for middle freq. range
                 42,  48,  54,  60,  66,
                 72,  78,  84,  90,  96,
                102, 108,
                120, 132, 156, 180 );     // wide interval for high freq. range

  // If the sampling rate is (standard) 44100Hz then the frequency of FFTData[x]
  // is 44100 / 512 * x, so the representative(or maximum) frequency of each
  // band is as followings
  //  BandOut[0] : 86.1Hz     BandOut[1] : 172.3Hz    BandOut[2] : 344.5Hz
  //  BandOut[3] : 516.8Hz    BandOut[4] : 1033.6Hz   BandOut[5] : 1550.4Hz
  //  . . . (516.8Hz interval upto BandOut[20]) . .   BandOut[20] : 9302.2Hz
  //  BandOut[21] : 10336Hz   BandOut[22] : 11370Hz   BandOut[23] : 13437Hz
  //  BandOut[24] : 15503Hz   . . very high frequency range is excluded

   Boost = 0.15; // Factor to intensify high frequency bands
   Scale = 80;   // Factor to calibrate the output range of BandOut[x]
                 //   ( approximately from 0 to 24 ).
var
   FFTData : TFFTData;
   NewBandOut : TBandOut;
   StartIndex : integer;
   i, k : integer;
   tmpIntensity : double;

begin
   if BASS_ChannelIsActive(Channel) <> BASS_ACTIVE_PLAYING then
      if BASS_ChannelIsActive(Channel) <> BASS_ACTIVE_STALLED then
      begin
         TimerFFT.Enabled := false;
         exit;
      end;

   if BASS_ChannelGetData(Channel, @FFTData, BASS_DATA_FFT512) = $ffffffff then
   begin
      for i := 0 to (NumFFTBands - 1) do
         BandOut[i] := 0;
      exit;
   end;

   TimerFFT.Enabled := false;  // Avoid duplicate timer interrupt

   for i := 0 to (NumFFTBands - 1) do
   begin
       if i = 0 then
          StartIndex := 1
       else
          StartIndex := FreqCoef[i-1] + 1;

    // tmpIntensity <= the greatest value in classified FFTData[x]'s
       tmpIntensity := 0;
       for k := StartIndex to FreqCoef[i] do
           if FFTData[k] > tmpIntensity then
              tmpIntensity := FFTData[k];

       NewBandOut[i] := round(tmpIntensity * (1 + i * Boost) * Scale);

    // Decrease the value of BandOut[i] smoothly for better looking
       if NewBandOut[i] > BandOut[i] then
          BandOut[i] := NewBandOut[i]
       else
          if BandOut[i] >= 2 then
             dec(BandOut[i], 2)
          else
             BandOut[i] := 0;

    // for the case that original NewBandOut[i] is smaller than BandOut[i] by 1
       if NewBandOut[i] > BandOut[i] then
          BandOut[i] := NewBandOut[i];
   end;

 //  if not GoVisOut then
      PostMessage(MessageHandle, WM_NewFFTData, 0, 0);

   if not GoVisOut then
   begin
      TimerFFT.Enabled := true;
      exit;
   end;

   VisThread.SetPosition(CurrentPosition);

  // Following sentences to rendering FFT data to visualization plug-ins are
  // executed in a different class to raise refresh rate
 {  p1 := ShareMemPointer;
   inc(p1, 280); // FFT Data region is allocated from offset 280 thru 2595 in shared memory
   p1^ := FstreamInfo.SampleRate; // BassChannelInfo.freq;
   inc(p1, 1);
   p1^ := FstreamInfo.Channels;   // BassChannelInfo.chans;  // * Added at Ver 1.4
   inc(p1, 1);
   p1^ := CurrentPosition;        // * Added at Ver 1.4
   inc(p1, 1);
   p2 := pointer(p1);

   FillChar(SpectrumData, SizeOf(SpectrumData), 0);
   FillChar(WaveformData, SizeOf(waveformData), 0);

   if FstreamInfo.Channels = 1 then
   begin
      Bass_ChannelGetdata(Channel, @spectrumdata, BASS_DATA_FFT2048);
      for i := 1 To 576 do
      begin
        p2^ := trunc(SpectrumData[i] * FVisScale);
        inc(p2);
      end;
      for i := 1 To 576 do
      begin
        p2^ := trunc(SpectrumData[i] * FVisScale);
        inc(p2);
      end;
   end else
 //  * Changed at Ver 1.4
 // Add BASS_DATA_FFT_INDIVIDUAL flag for seperate FFT per each channel
   begin
      Bass_ChannelGetdata(Channel, @spectrumdata, BASS_DATA_FFT2048 + BASS_DATA_FFT_INDIVIDUAL);
      for i := 1 To 576 do
      begin
        p2^ := trunc(SpectrumData[i*2] * FVisScale);
        inc(p2);
      end;
      for i := 1 To 576 do
      begin
        p2^ := trunc(SpectrumData[i*2 + 1] * FVisScale);
        inc(p2);
      end;
   end;

   if FstreamInfo.Channels = 1 then
   begin
     Bass_ChannelGetdata(Channel, @waveformdata, 1152);
     for i := 0 to 575 do
     begin
       p2^ := waveformdata[i] div 256;
       inc(p2);
     end;
     for i := 0 to 575 do
     begin
       p2^ := waveformdata[i] div 256;
       inc(p2);
     end;
   end else
   begin
      Bass_ChannelGetdata(Channel, @waveformdata, 2304);
      for i := 0 to 575 do
      begin
        p2^ := waveformdata[i*2] div 256;
        inc(p2);
      end;
      for i := 0 to 575 do
      begin
        p2^ := waveformdata[i*2+1] div 256;
        inc(p2);
      end;
   end;

 // Inform "VisOut.exe" that visualization data is ready
   fisSharedMemory1.InformDataReady(DataReady, 2316); }

   TimerFFT.Enabled := true;
end;


procedure TBASSPlayer.ShowPluginConfigForm;
var
   SearchRec : TSearchRec;
begin
 // How do we catch the values of parent's elements in the program of child Form ?
 // I cannot implement to get the values of TBASSPlayer's elements in the program
 // of PluginConfigForm.OnShow event procedure.
 // So I have to write the program in the way that TBASSPlayer hands over the
 // values of its elements to PluginConfigForm's.
   if not FBASSReady then
   begin
      Application.MessageBox('Player is not ready !', 'Confirm', MB_OK);
      exit;
   end;

   if FPluginFirst then
      PluginConfigForm.PluginFirst.Checked := true
   else
      PluginConfigForm.PluginFirst.Checked := false;

   PluginConfigForm.PluginList.Clear;
   if FindFirst(GetProgDir + 'Plugins\in_*.dll', 0, SearchRec) = 0 then
   begin
      PluginConfigForm.PluginList.Items.Add(SearchRec.Name);
      while FindNext(SearchRec) = 0 do
         PluginConfigForm.PluginList.Items.Add(SearchRec.Name);
   end;
   FindClose(SearchRec);

   with PluginConfigForm.PluginGrid do
   begin
      Cells[0, 0] := 'Name';
      Cells[1, 0] := 'Decsription';
      Cells[2, 0] := 'Playable File Types';
   end;
   PluginConfigForm.ClearPluginGrid;
   PluginConfigForm.FillPluginGrid;
   SetMessageHandle(MessageHandle);
   PluginConfigForm.ShowModal;
end;

procedure TBASSPlayer.HideVisWindow;
begin
   if not GoVisOut then
      exit;

   VisThread.fisSharedMemory1.InformDataReady(MinimizeWindow, 0{Data Size});
end;

procedure TBASSPlayer.ShowVisWindow;
begin
   if not GoVisOut then
      exit;

   VisThread.fisSharedMemory1.InformDataReady(RestoreWindow, 0{Data Size});
end;

function TBASSPlayer.GetMessageHandle : HWND;
begin
   result := MessageHandle;
end;

// Inform current player status to visualizer
procedure TBASSPlayer.InformPlayerStatus(StatusIndex : DWORD);   // * Added at Ver 1.5
var
   p1 : PDWORD;
   p2 : PBYTE;
   ModeId : DWORD;
   TitleBuf : array[0..255] of char;

// The shared memory obtained by fisSharedMemory1 is used as follows
//   offset   0 -    3 : Player Mode
//            4 -  275 : Stream Information
//          276 -  279 : SyncWindows flag
//          280 -  283 : Flag to notify that new FFT data is given
//          284 -  287 : Sample rate
//          288 -  291 : number of channels
//          292 -  295 : playback position;
//          296 - 2599 : FFT Data

begin
   if not GoVisOut then
      exit;

   TimerFFT.Enabled := false;

   p1 := VisThread.ShareMemPointer;

   case StatusIndex of
      stPlayerMode : begin
           ModeId := ord(FPlayerMode);
           p1^ := ModeId;
           VisThread.fisSharedMemory1.InformDataReady(InformPlayerMode, 4);
         end;
      stStreamInfo : begin
           inc(p1, 4);
           p1^ := FStreamInfo.SampleRate;
           inc(p1, 1);
           p1^ := FStreamInfo.BitRate;
           inc(p1, 1);
           p1^ := FStreamInfo.Channels;
           inc(p1, 1);
           p1^ := FStreamInfo.Duration;
           inc(p1, 1);
           FillChar(TitleBuf, 256, 0);
           StrPCopy(TitleBuf, FStreamInfo.Title);
           p2 := pointer(p1);
           move(TitleBuf, p2^, 256);
           VisThread.fisSharedMemory1.InformDataReady(InformStreamInfo, 272);
         end;
      stSyncWindows : begin
           inc(p1, 276);
           if FSyncVisWindow then
              p1^ := 1
           else
              p1^ := 0;
           VisThread.fisSharedMemory1.InformDataReady(InformSyncWindows, 4);
         end;
   end;

   TimerFFT.Enabled := true;
end;


procedure Register;
begin
   RegisterComponents('Samples', [TBASSPlayer]);
end;

end.
