// Unit BASSTest
//   The program to test TBASSPlayer commponent
//
//    written by Silhwan Hyun  hsh@chollian.net
//
// Note : bass.dll & basswma.dll should be copied in the directory where executable
//        (compiled) program resides in and Winamp 2 input plug-ins in the
//        directory <directory where executable program resides in>\Plugins
//
// Ver 1.5                         15 Dec 2003
//  - Added components and codes to test the functions/properties for playing internet stream.
//  - Added 'DSP plug-in Control' button to set up Winamp DSP plug-in
//  - Changed output volume range : 0 ~ 100 -> 0 ~ 255
//
// Ver 1.4                         12 May 2003
//  - Added message handling to set up synchronized action between the windows
//    of BassPlay and the one of Winamp 2 visualization plug-in
//
// Ver 1.3                         5 Apr 2003
//  - Added a button to invoke Winamp visualization plug-in control form, which
//    enables you to select, start and stop visualization plug-ins.
//  - Added two sliders to adjust the depth of echo & reverb effects
//
// Ver 1.2                         5 Feb 2003
//  - Added a button to invoke Winamp input plug-in configuration form to play
//    stream files which BASS sound system cannot decode.
//  - Compiled with Delphi 4
//
// Ver 1.1                         8 Dec 2002
//   - Compiled on Delphi 4 for old version Delphi users
//   - Modified procedure DisplayFFTBand for better result
//
// Ver 1.0                         4 Aug 2002
//   - Initial release
//   - Compiled with Delphi 6

unit BassTest;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, BassPlayer, PluginCtrl, Knob, slider, ExtCtrls,
  ComCtrls;

type
  TMainForm = class(TForm)
    btnOpen: TButton;
    OpenDialog1: TOpenDialog;
    btnPlay: TButton;
    btnPause: TButton;
    btnStop: TButton;
    btnClose: TButton;
    VolumeKnob: TKnob;
    PosSlider: TSlider;
    cbFlanger: TCheckBox;
    cbEcho: TCheckBox;
    cbReverb: TCheckBox;
    Slider2: TSlider;
    Slider3: TSlider;
    Slider4: TSlider;
    Slider5: TSlider;
    Slider6: TSlider;
    Slider7: TSlider;
    Slider8: TSlider;
    Slider9: TSlider;
    cbEqualizer: TCheckBox;
    Timer_Stat: TTimer;
    Label_State: TLabel;
    Label_Length: TLabel;
    Slider1: TSlider;
    Slider10: TSlider;
    Label1: TLabel;
    Label2: TLabel;
    Panel1: TPanel;
    GaugePaintBox: TPaintBox;
    Label3: TLabel;
    Label_Title: TLabel;
    InfoMemo: TMemo;
    btnPluginSetup: TButton;
    EchoSlider: TSlider;
    ReverbSlider: TSlider;
    btnShowVisCtrlForm: TButton;
    btnShowDSPCtrlForm: TButton;
    Label6: TLabel;
    edBitrate: TEdit;
    Panel2: TPanel;
    ComboBox1: TComboBox;
    Label5: TLabel;
    ProgressBar1: TProgressBar;
    btnNetOpen: TButton;
    Label7: TLabel;

    procedure FormShow(Sender: TObject);
    function  OpenStream(StreamName : string) : boolean;
    procedure btnOpenClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btnPlayClick(Sender: TObject);
    procedure btnPauseClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure cbFlangerClick(Sender: TObject);
    procedure cbEchoClick(Sender: TObject);
    procedure cbReverbClick(Sender: TObject);
    procedure VolumeKnobChange(Sender: TObject);
    procedure EQSliderChange(Sender: TObject);
    procedure cbEqualizerClick(Sender: TObject);
    procedure PosSliderStartTracking(Sender: TObject);
    procedure PosSliderStopTracking(Sender: TObject);
    procedure Timer_StatTimer(Sender: TObject);
    procedure CreateBasicImage;
    procedure ShowBackground;
    procedure PluginStartPlay(Sender: TObject; ChannelInfo : PChannelInfo);
    procedure PluginChannelInfo(Sender: TObject; ChannelInfo : PChannelInfo);
    procedure DisplayFFTBand(Sender: TObject; Bands : TBandOut);
    procedure GetMETAFromNet(Sender: TObject; Content : string);
    procedure DownloadEnded(Sender: TObject; Content : string);
    procedure btnPluginSetupClick(Sender: TObject);
    procedure EchoSliderChange(Sender: TObject);
    procedure ReverbSliderChange(Sender: TObject);
    procedure btnShowVisCtrlFormClick(Sender: TObject);
    procedure btnShowDSPCtrlFormClick(Sender: TObject);
    procedure btnNetOpenClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure WndProc(var Msg : TMessage); override;
  end;

var
  MainForm: TMainForm;
  BassPlayer1 : TBassPlayer;

implementation

uses VisPluginCtrl, DSPPluginCtrl;


{$R *.dfm}

const
// Constants for frequency spectrum visualizzer
   BlockWidth = 3;
   HBlockGap = 1;
   HBlockCount = NumFFTBands;
   VLimit = 24;    // Intensity range : 0 ~ 24
   BackColor = clBlack;

var
   SPS : LongInt;    // sample rate
   NowTracking : boolean;
   BasicBMP : TBitMap;
   GaugeTempBMP : TBitMap;
   DisplayBar: TBitmap;
   GaugeRect : TRect;
   EQGains : TEQGains;
   PeakValue : array[1..NumFFTBands] of single;
   PassedCounter : array[1..NumFFTBands] of integer;
   IsNetRadio : boolean;
   IsNetStream : boolean;

procedure TMainForm.WndProc(var Msg : TMessage);
var
   WinHwnd : HWND;
   BaseWindowHandle : HWND;
   VisWindowHandle : HWND;
   ProcId1, ProcId2 : DWORD;
   ClassName : array[0..255] of Char;
begin
  // Because the visualizer runs in a seperate process there is no automatic
  // synchronized action between the windows of BassPlay and visualizer
  // when BassPlay(or visualizer) gets focus, hides or shows its windows.
  // So, I added following message handling to set up corelation between them.
  // (visualizer = Winamp visualization plug-in or its visible window)

  // message WM_ACTIVATEAPP is received when BassPlay gets focus or loses focus.
   if (Msg.Msg = WM_ACTIVATEAPP) then
   begin
    // Msg.WParam = 0 : BassPlay losed focus
    // BassPlayer1.SyncVisWindow = false : synchronized action is disabled
      if (Msg.WParam = 0) or (not BassPlayer1.SyncVisWindow) then
      begin
         inherited WndProc(Msg);
         exit;
      end;

  // Put the visualizer just below the BassPlay's windows when BassPlay gets focus.

      VisWindowHandle := BassPlayer1.GetVisWindowHandle;  // VisWindowHandle = the handle of visualizer
      if VisWindowHandle = 0 then    // if there is no visualizer
      begin
         inherited WndProc(Msg);
         exit;
      end;

      GetWindowThreadProcessId(Self.Handle, @ProcId1);  // get the process Id. of BassPlay

    // get the handle of a window laid just above visualizer
      BaseWindowHandle := VisWindowHandle;
      repeat
         WinHwnd := GetWindow(BaseWindowHandle, GW_HWNDPREV);
         BaseWindowHandle := WinHwnd;             // for next iteration
         if WinHwnd <> 0 then
            if IsWindowVisible(WinHwnd) then      // neglect if invisible window
               if not IsIconic(WinHwnd) then      // neglect if minimized window
               begin
                  GetClassName(WinHwnd, ClassName, SizeOf(ClassName));
                  if string(ClassName) <> 'TApplication' then
                     break;
               end;
      until (WinHwnd = 0);

      if WinHwnd <> 0 then
      begin
         GetWindowThreadProcessId(WinHwnd, @ProcId2);
         if ProcId1 = ProcId2 then   // the upper window is the one of BassPlay ?
         begin
            inherited WndProc(Msg);
            exit;
         end;
      end else
      begin
         inherited WndProc(Msg);
         exit;
      end;

    // get the handle of the lowest window among the BassPlay's windows
      repeat
         BaseWindowHandle := WinHwnd;       // for next iteration
         WinHwnd := GetWindow(BaseWindowHandle, GW_HWNDPREV); // get the upper window's handle
         if WinHwnd = 0 then
            break
         else
            if IsWindowVisible(WinHwnd) then
               if not IsIconic(WinHwnd) then
               begin
                  GetClassName(WinHwnd, ClassName, SizeOf(ClassName));
                  if string(ClassName) <> 'TApplication' then
                     GetWindowThreadProcessId(WinHwnd, @ProcId2); // get process Id. for the upper window
               end;
      until (ProcId1 = ProcId2); // the upper window is the one of BassPlay's ?

   // put the visualizer just below the BassPlay's
      if WinHwnd <> 0 then
         SetWindowPos(VisWindowHandle, WinHwnd, 0, 0, 0, 0,
                                       SWP_NOACTIVATE + SWP_NOMOVE + SWP_NOSIZE);
      Msg.result := 0;
   end else

 // Show or hide visualizer when BassPlay is showing or hiding
   if (Msg.Msg = WM_SHOWWINDOW) then
   begin
      VisWindowHandle := BassPlayer1.GetVisWindowHandle;
      if (VisWindowHandle = 0) or (not BassPlayer1.SyncVisWindow) then    // VisWindowHandle = 0 : if there is no visualizer
      begin
         inherited WndProc(Msg);
         exit;
      end;

      if Msg.wParam = 0 then         // = BassPlay is hiding
         BASSPlayer1.HideVisWindow   // hide visualizer
      else if IsIconic(VisWindowHandle) then
         BASSPlayer1.ShowVisWindow;  // show visualizer

      Msg.result := 0;
   end;

   inherited WndProc(Msg);
end;

procedure TMainForm.FormShow(Sender: TObject);
var
   i : integer;
begin
   BassPlayer1 := TBassPlayer.Create(Self);
   if BassPlayer1.BASSDLLVer = '' then
   begin
      Application.MessageBox('BASS.DLL is not loaded.', 'Confirm', MB_OK);
      exit;
   end;
   if not BassPlayer1.BASSWMAReady then
      Application.MessageBox('BASSWMA.DLL is not loaded.', 'Confirm', MB_OK);

   BassPlayer1.OnNewFFTData := DisplayFFTBand;
   BassPlayer1.OnPluginStartPlay := PluginStartPlay;
   BassPlayer1.OnGetChannelInfo := PluginChannelInfo;
   BassPlayer1.OnGetMeta := GetMETAFromNet;
   BassPlayer1.OnDownloaded := DownloadEnded;

   VolumeKnob.Position := BassPlayer1.Volume;
   CreateBasicImage;
   ShowBackground;
   for i := 1 to NumFFTBands do
      PeakValue[i] := 0;

   EQGains[0] := Slider1.Value - 15.0;
   EQGains[1] := Slider2.Value - 15.0;
   EQGains[2] := Slider3.Value - 15.0;
   EQGains[3] := Slider4.Value - 15.0;
   EQGains[4] := Slider5.Value - 15.0;
   EQGains[5] := Slider6.Value - 15.0;
   EQGains[6] := Slider7.Value - 15.0;
   EQGains[7] := Slider8.Value - 15.0;
   EQGains[8] := Slider9.Value - 15.0;
   EQGains[9] := Slider10.Value - 15.0;

   EchoSlider.Value := BassPlayer1.EchoLevel;
   ReverbSlider.Value := BassPlayer1.ReverbLevel;

   if BassPlayer1.CanUsePlugin then
   begin
      LoadWinampPlugin('in_cdda.dll');
      LoadWinampPlugin('in_asf.dll');
      LoadWinampPlugin('in_midi.dll');
   // The 'in_wm_old.dll' is needed to play the WMA files on the internet.
   //  note) The original name of 'in_wm_old.dll' is 'in_wm.dll'.
   //        I renamed the old version (v1.00) of 'in_wm.dll' to 'in_wm_old.dll'
   //        to differentiate from recent versions of 'in_wm.dll'.
   // Because recent versions of 'in_wm.dll' does not output decoded sound data,
   // we cannot get FFT data for spectrum display and cannot apply DSP effects
   // if we use them.
      LoadWinampPlugin('in_wm_old.dll');
   end;

   ComboBox1.ItemIndex := 0;

   if BassPlayer1.PlayerReady then
      Timer_Stat.Enabled := true;
end;

function TMainForm.OpenStream(StreamName : string) : boolean;
var
   s1, s2 : string;
   URLInfo : pchar;
begin
   result := false;
   IsNetRadio := false;
   IsNetStream := false;

   if BassPlayer1.Open(StreamName) then
   begin
      Label_Title.Caption := BassPlayer1.StreamInfo.Title;
      Label_Length.Caption := 'Length ' +
            FormatDateTime ('nn:ss', BassPlayer1.PlayLength / (1000 * 24 * 60 * 60));
      InfoMemo.Clear;
      SPS := 0;
      edBitrate.Text := '';

      if BassPlayer1.IsNetRadio then
      begin
         InfoMemo.Lines.Add('Connected to an internet radio station.');

         URLInfo := BassPlayer1.ICYInfo;
         if URLInfo <> nil then
            while StrLen(URLInfo) > 0 do
            begin
               if pos('icy-name:', string(URLInfo)) <> 0 then
               begin
                  inc(URLInfo, 9);
                  InfoMemo.Lines.Add('icy-name : ' + string(URLInfo));
                  break;
               end;
               inc(URLInfo, StrLen(URLInfo) + 1);
            end;

         if BassPlayer1.StreamInfo.BitRate > 0 then
            edBitrate.Text := intToStr(BassPlayer1.StreamInfo.BitRate) + 'KBPS';
         IsNetRadio := true;
      end else
      if BassPlayer1.IsNetStream then
      begin
         SPS := BassPlayer1.StreamInfo.SampleRate;
         if SPS <> 0 then
         begin
            InfoMemo.Lines.Add('Playing an internet stream.');
            edBitrate.Text := intToStr(BassPlayer1.StreamInfo.BitRate) + 'KBPS';
         end;
         URLInfo := BassPlayer1.HTTPInfo;
         if URLInfo <> nil then
            while StrLen(URLInfo) > 0 do
            begin
               if pos('Server:', string(URLInfo)) <> 0 then
               begin
                  InfoMemo.Lines.Add(string(URLInfo));
                  break;
               end;
               inc(URLInfo, StrLen(URLInfo) + 1);
            end;
         IsNetStream := true;


        // <-
         URLInfo := BassPlayer1.ICYInfo;
         if URLInfo <> nil then
            while StrLen(URLInfo) > 0 do
            begin
               InfoMemo.Lines.Add(string(URLInfo));
               inc(URLInfo, StrLen(URLInfo) + 1);
            end;   // <-



      end else   // local stream file
      begin
         InfoMemo.Lines.Add('Playing ' + ExtractFileName(BassPlayer1.StreamInfo.FileName));
         if BassPlayer1.DecodingByPlugin then
            InfoMemo.Lines.Add(' (Being decoded by Winamp plug-in)');
         edBitrate.Text := intToStr(BassPlayer1.StreamInfo.BitRate) + 'KBPS';
      end;

      if not IsNetStream then
         SPS := BassPlayer1.StreamInfo.SampleRate;
      if SPS <> 0 then
      begin
         s1 := intToStr(SPS) + 'Hz';
         if BassPlayer1.StreamInfo.Channels = 1 then
            S2 := 'Mono'
         else
            s2 := 'Stereo';
         InfoMemo.Lines.Add(s1 + ', ' + s2);
      end;

   // Following statements are only to scroll down Memo1's contents, i.e.,
   //  to make Memo1's 1st line as the top line.
      InfoMemo.SetFocus;
      InfoMemo.SelStart := 0;
      InfoMemo.SelLength := 0;
      BassPlayer1.Play;

      result := true;
   end;
end;

procedure TMainForm.btnOpenClick(Sender: TObject);
var
   s1, s2 : string;
begin
   if not BassPlayer1.PlayerReady then
   begin
      Application.MessageBox('Cannot open file (BassPlayer is not ready)',
                 'Confirm', MB_OK);
      exit;
   end;

   OpenDialog1.FileName := '';
   s1 := BassPlayer1.NativeFileExts;
   s2 := BassPlayer1.PluginFileExts;
   OpenDialog1.Filter := 'All playable files |' + s1 + s2 + '|' +
                         'BASS native files (' + s1 + ')|' + s1 + '|' +
                         'Plug-in supported files (' + s2 + ')|' + s2;

   if OpenDialog1.Execute then
      OpenStream(OpenDialog1.FileName);
end;


procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   BassPlayer1.QuitVisoutProgram;
   BassPlayer1.Free;
   BasicBMP.Free;
   GaugeTempBMP.Free;
   DisplayBar.Free;
end;

procedure TMainForm.btnPlayClick(Sender: TObject);
begin
   BassPlayer1.Play;
  { if BassPlayer1.Mode = plmPlaying then
      Timer_Stat.Enabled := true;  }
end;

procedure TMainForm.btnPauseClick(Sender: TObject);
begin
   BassPlayer1.Pause(true);
end;

procedure TMainForm.btnStopClick(Sender: TObject);
begin
   BassPlayer1.Stop;
end;

procedure TMainForm.btnCloseClick(Sender: TObject);
begin
   Close;
end;

procedure TMainForm.cbFlangerClick(Sender: TObject);
begin
   if cbFlanger.Checked then
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects + [Flanger]
   else
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects - [Flanger];

end;

procedure TMainForm.cbEchoClick(Sender: TObject);
begin
   if cbEcho.Checked then
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects + [Echo]
   else
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects - [Echo];
end;

procedure TMainForm.cbReverbClick(Sender: TObject);
begin
   if cbReverb.Checked then
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects + [Reverb]
   else
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects - [Reverb];
end;


procedure TMainForm.cbEqualizerClick(Sender: TObject);
begin
   if cbEqualizer.Checked then
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects + [Equalizer]
   else
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects - [Equalizer];
end;

procedure TMainForm.VolumeKnobChange(Sender: TObject);
begin
   BassPlayer1.Volume := VolumeKnob.Position;
end;

procedure TMainForm.EQSliderChange(Sender: TObject);
var
   BandNum : integer;
begin
   BandNum := (Sender as TSlider).Tag - 1;
   EQGains[BandNum] := (Sender as TSlider).Value - 15.0;
   BassPlayer1.EQGains := EQGains;
end;


procedure TMainForm.PosSliderStartTracking(Sender: TObject);
begin
   NowTracking := true;
end;

procedure TMainForm.PosSliderStopTracking(Sender: TObject);
var
   SongPos : int64;
begin
   if BassPlayer1.Seekable then
   begin
      SongPos := Trunc(PosSlider.Value * BassPlayer1.PlayLength / PosSlider.MaxValue);
      BassPlayer1.Position := SongPos;
   end;

   NowTracking := false;
end;

procedure TMainForm.Timer_StatTimer(Sender: TObject);
begin
   if BassPlayer1.Mode = plmStandby then
      Label_State.Caption := 'Standby'
   else if BassPlayer1.Mode = plmReady then
      Label_State.Caption := 'Ready to play'
   else if BassPlayer1.Mode = plmPlaying then
      Label_State.Caption := 'Playing '
                 +  FormatDateTime ('nn:ss', BassPlayer1.Position / (1000 * 24 * 60 * 60))
   else if BassPlayer1.Mode = plmStopped then
      Label_State.Caption := 'Stopped'
   else if BassPlayer1.Mode = plmPaused then
      Label_State.Caption := 'Paused';

   if BassPlayer1.Mode <> plmPlaying then
   begin
   //   Timer_Stat.Enabled := false;
      ShowBackground;
   end;

   if BassPlayer1.Mode = plmStandby then
      exit;

   if not IsNetRadio then
      if not NowTracking and (BassPlayer1.Mode <> plmStopped) then
         if BassPlayer1.PlayLength > 0 then
            PosSlider.Value := (BassPlayer1.Position * PosSlider.MaxValue) div BassPlayer1.PlayLength;
   if IsNetStream then
   begin
      if BassPlayer1.DecodingByPlugin then
         ProgressBar1.Position := 0  // not available if decoded by Winamp plug-in
      else
         ProgressBar1.Position := round((BassPlayer1.DownloadProgress * ProgressBar1.Max)
                                      / BassPlayer1.StreamInfo.FileSize)
   end else
      ProgressBar1.Position := 0;
end;


// Set background image & bar image for spectrum display
procedure TMainForm.CreateBasicImage;
var
   i : integer;
   R, G, B: Integer;
begin
   BasicBMP := TBitMap.Create;
   BasicBMP.Width := (BlockWidth + HBlockGap) * HBlockCount - HBlockGap + 3;
   BasicBMP.Height := VLimit + 1;
   GaugePaintBox.Width := BasicBMP.Width;
   GaugePaintBox.Height := BasicBMP.Height;
   GaugeTempBMP := TBitMap.Create;
   GaugeTempBMP.Width := BasicBMP.Width;
   GaugeTempBMP.Height := BasicBMP.Height;
   GaugeRect.Left := 0;
   GaugeRect.Top := 0;
   GaugeRect.Right := BasicBMP.Width;
   GaugeRect.Bottom := BasicBMP.Height;

   with BasicBMP.Canvas do
   begin
      Brush.Color := BackColor;
      Brush.Style := bsSolid;
      FillRect(Rect(0, 0, BasicBMP.Width, BasicBMP.Height));
      Pen.Color := clBlue;
      Pen.Width := 1;
      Pen.Style := psDot;
      MoveTo(0, 0);
      LineTo(0, BasicBMP.Height);        // Draw Y-axis line
      MoveTo(0, BasicBMP.Height - 1);
      LineTo(BasicBMP.Width, BasicBMP.Height - 1); // Draw X-axis line
   end;

   DisplayBar := TBitmap.Create;
   DisplayBar.PixelFormat := pf32bit;
   DisplayBar.Width := BlockWidth;
   DisplayBar.Height := VLimit;

   R := 255;
   G := 0;
   B := 0;

   for i := 0 to VLimit - 1 do
   begin
     if i > VLimit / 2 then
        Dec(R, Trunc(256 / VLimit))
     else
        Inc(G, Trunc(768 / VLimit));
     if R < 0 then R := 0;
     if G > 255 then G := 255;
     DisplayBar.Canvas.Brush.Color := TColor(RGB(R, G, B));
     DisplayBar.Canvas.FillRect(Rect(0, i, BlockWidth, i + 1));
   end;

end;

// Winamp plug-in notified that it has started playing a stream file
procedure TMainForm.PluginStartPlay(Sender: TObject; ChannelInfo : PChannelInfo);
var
   s1, s2 : string;
begin
   if SPS = 0 then     // = have not gotten stream information yet
   begin
      if IsNetStream then
      begin
         InfoMemo.Lines.Add('Playing an internet stream.');
         if BassPlayer1.DecodingByPlugin then
            InfoMemo.Lines.Add(' (Being decoded by Winamp plug-in)');
      end;

      SPS := ChannelInfo^.SampleRate;
      if SPS <> 0 then
      begin
         s1 := intToStr(SPS) + 'Hz';
         if ChannelInfo^.Channels = 1 then
            S2 := 'Mono'
         else
            s2 := 'Stereo';
         InfoMemo.Lines.Add(s1 + ', ' + s2);
      end;

      InfoMemo.SetFocus;
      InfoMemo.SelStart := 0;
      InfoMemo.SelLength := 0;

      Label_Length.Caption := 'Length ' +
         FormatDateTime ('nn:ss', BassPlayer1.PlayLength / (1000 * 24 * 60 * 60));
      Label_Title.Caption := BassPlayer1.StreamInfo.Title;
   end;
end;

// Winamp plug-in has notified that it had gotten channel information on a playing stream file
// The first call of this procedure precedes PluginStartPlay.
procedure TMainForm.PluginChannelInfo(Sender: TObject; ChannelInfo : PChannelInfo);
begin
  // SPS := ChannelInfo^.SampleRate;
   edBitrate.Text := intToStr(ChannelInfo^.BitRate) + 'KBPS';
end;

procedure TMainForm.GetMETAFromNet(Sender: TObject; Content : string);
begin
   Label_Title.Caption := Content;
end;

procedure TMainForm.DownloadEnded(Sender: TObject; Content : string);
begin
 // Now we can get the exact value of playback length of an internet stream.
   Label_Length.Caption := 'Length ' +
            FormatDateTime ('nn:ss', StrToInt(Content) / (1000 * 24 * 60 * 60));
            
 // We may have gotten correct 'Title' for playing stream file.
   if (BassPlayer1.StreamInfo.Title <> '') then
      Label_Title.Caption := BassPlayer1.StreamInfo.Title;
end;

// Display spectrum image
procedure TMainForm.DisplayFFTBand(Sender: TObject; Bands : TBandOut);
var
   tmpRect, BarRect : TRect;
   j : integer;
begin
 // To prevent flickering, use temporary image buffer and process as follows
 // 1) Draw output image on temporary image buffer.
 // 2) Copy temporary image buffer's image to display canvas

 // Copy BasicBMP's image to GaugeTempBMP(used as temporary image buffer)
   BitBlt(GaugeTempBMP.Canvas.Handle, // handle to destination device context
          GaugeRect.Left,	// x-coordinate of destination rectangle's upper-left corner
          GaugeRect.Top,	// y-coordinate of destination rectangle's upper-left corner
          GaugeTempBMP.Width,	// width of destination rectangle
          GaugeTempBMP.Height,	// height of destination rectangle
          BasicBMP.Canvas.Handle, // handle to source device context
          GaugeRect.Left,	// x-coordinate of source rectangle's upper-left corner
          GaugeRect.Top,	// y-coordinate of source rectangle's upper-left corner
          SRCCOPY);             // Copies the source rectangle directly to the destination rectangle.
 // Substituted Canvas.CopyRect with BitBlt for speed up
 // GaugeTempBMP.Canvas.CopyRect(GaugeRect, BasicBMP.Canvas, GaugeRect);

 // Draw spectrum image to GaugeTempBMP
   for j := 1 to HBlockCount do
   begin
      if Bands[j-1] > VLimit then
         Bands[j-1] := VLimit;

      if Bands[j-1] > 0 then
      begin
     // Copy partial image of DisplayBar to GaugeTempBMP
        BarRect.Left := 0;
        BarRect.Right := BlockWidth;
        BarRect.Top := VLimit - Bands[j-1];
        if BarRect.Top < 0 then
           BarRect.Top := 0;
        BarRect.Bottom := DisplayBar.Height;

        tmpRect.Left := (BlockWidth + HBlockGap) * (j - 1) + 2;
        tmpRect.Right := tmpRect.Left + BlockWidth;
        tmpRect.Top := BarRect.Top;
        tmpRect.Bottom := BarRect.Bottom;

        BitBlt(GaugeTempBMP.Canvas.Handle,
               tmpRect.Left,
               tmpRect.Top,
               BlockWidth,
               tmpRect.Bottom - tmpRect.Top + 1,
               DisplayBar.Canvas.Handle,
               BarRect.Left,
               BarRect.Top,
               SRCCOPY);
      end;

      if Bands[j-1] >= trunc(PeakValue[j]) then
      begin
         PeakValue[j] := Bands[j-1] + 0.01;  // 0.01 : to compensate round off
         PassedCounter[j] := 0;
      end else if Bands[j-1] < trunc(PeakValue[j]) then
      begin
         if trunc(PeakValue[j]) > 0 then
         begin
            with GaugeTempBMP.Canvas do
            begin
            // Draw peak line
               Pen.Color := TColor(RGB(192, 192, 192));   // color for peak line
               MoveTo((BlockWidth + HBlockGap) * (j - 1) + 2, VLimit - trunc(PeakValue[j]));
               LineTo((BlockWidth + HBlockGap) * (j - 1) + 2 + BlockWidth, VLimit - trunc(PeakValue[j]));
            end;

      // Followings are to show simillar spectrum image to WINAMP's
      //  - Put delay time before lowering peak line
      //  - Accerate lowering speed according to the time elapsed
            if PassedCounter[j] >= 8 then
                PeakValue[j] := PeakValue[j] - 0.3 * (PassedCounter[j] - 8);

            if PeakValue[j] < 0 then
               PeakValue[j] := 0
            else
               inc(PassedCounter[j]);
         end;
      end;

   end;

 // Copy GaugeTempBMP's image to GaugePaintBox
   BitBlt(GaugePaintBox.Canvas.Handle,
          GaugeRect.Left,
          GaugeRect.Top,
          GaugeTempBMP.Width,
          GaugeTempBMP.Height,
          GaugeTempBMP.Canvas.Handle,
          GaugeRect.Left,
          GaugeRect.Top,
          SRCCOPY);
end;

// Show background image during non-playing period
procedure TMainForm.ShowBackground;
begin
   BitBlt(GaugePaintBox.Canvas.Handle,
          GaugeRect.Left,
          GaugeRect.Top,
          BasicBMP.Width,
          BasicBMP.Height,
          BasicBMP.Canvas.Handle,
          GaugeRect.Left,
          GaugeRect.Top,
          SRCCOPY);
end;


procedure TMainForm.btnPluginSetupClick(Sender: TObject);
begin
   BassPlayer1.ShowPluginConfigForm;
end;

procedure TMainForm.EchoSliderChange(Sender: TObject);
begin
   BassPlayer1.EchoLevel := EchoSlider.Value;
end;

procedure TMainForm.ReverbSliderChange(Sender: TObject);
begin
   BassPlayer1.ReverbLevel := ReverbSlider.Value;
end;

procedure TMainForm.btnShowVisCtrlFormClick(Sender: TObject);
begin
   VisControlForm.Show;
end;

procedure TMainForm.btnShowDSPCtrlFormClick(Sender: TObject);
begin
   DSPControlForm.Show;
end;

procedure TMainForm.btnNetOpenClick(Sender: TObject);
begin
   if not BassPlayer1.PlayerReady then
   begin
      Application.MessageBox('Cannot open file (BassPlayer is not ready)',
                 'Confirm', MB_OK);
      exit;
   end;

   OpenStream(ComboBox1.Text);
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
   ComboBox1.Items.Add('http://205.188.234.129:8024');
   ComboBox1.Items.Add('http://69.31.76.84:8000');
   ComboBox1.Items.Add('http://193.201.220.87:8000');
   ComboBox1.Items.Add('http://206.98.167.99:8006');
   ComboBox1.Items.Add('http://205.188.234.4:8016');
   ComboBox1.Items.Add('mms://217.11.251.151/rmffm');
   ComboBox1.Items.Add('http://server2.somafm.com:8000');
   ComboBox1.Items.Add('http://server2.somafm.com:8082');
   ComboBox1.Items.Add('http://user.chollian.net/~hsh/sound/Korean Song - A girl''s love story.asf');
   ComboBox1.Items.Add('http://user.chollian.net/~hsh/sound/Susan Jacks - Evergreen.wma');
   ComboBox1.Items.Add('http://user.chollian.net/~hsh/sound/Celine Dion - I owe you.mp3');
   ComboBox1.Items.Add('http://user.chollian.net/~hsh/sound/Bette Midler - The Rose.ogg');
   ComboBox1.Items.Add('http://user.chollian.net/~hsh/sound/-ȥڰƴѳ{ost}.mp3');
end;
    
end.
