// Main unit of VisOut.dpr
// The form of this unit is not visible at run time.
// It only takes charge of driving Winamp visualization plug-in.
//    written by Silhwan Hyun  hsh@chollian.net
//
//
// Ver 1.2                         15 Dec 2003
//   Added functions and codes to use some Winamp visualization plug-ins which
//   require that their cooperator should be the Winamp's main window.
//
// Ver 1.1                         12 May 2003
//   Added message handling to set up synchronized action between the windows
//    of BassPlay and the one of Winamp 2 visualization plug-in
//
// Ver 1.0                         5 Apr 2003
//   - Initial release

unit VisOutIntf;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ioplug, PluginCtrl, fisSharedMemory, ExtCtrls;

type
  TdumbForm = class(TForm)
    fisSharedMemory1: TfisSharedMemory;
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    { Private declarations }
    procedure RenderToVis;
    procedure EndProgram;
    procedure SetSyncWithMain;
    procedure HideVisWindow;
    procedure ShowVisWindow;
    procedure GetStatus(StatusId : DWORD);
  public
    { Public declarations }
    procedure HandleMessages(var Msg: TMsg; var Handled: Boolean);
  end;

  procedure PopUpMainWindow;

// Because some plug-ins require that their cooperator should be Winamp's main window,
// I added following functions for emulation of Winamp's main window
  function CreateFakeWindow : HWND;                   // * New at Ver 1.2
  function DestroyFakeWindow : boolean;               // * New at Ver 1.2

var
  dumbForm: TdumbForm;

implementation

{$R *.DFM}


var
   ShareMemPointer : pointer;
   DataReadyMsg : hwnd;

   PluginName : string;
   VisModNum : integer;
   Vismod : TVismod;
   NumVismod : integer;
   MainProcId : DWORD;
   VisProcId  : DWORD;
   VisWindowHandle : HWND = 0;
   VisualizerQuitted : boolean = false;

   MainWindowHandle : HWND;
   MsgHandle : HWND;
   VisOutHandle : HWND;
   OldWindowProc: Pointer;
   EndByProgram : boolean;
   SyncWithMain : boolean;
   ErrorMode : Word;

   wClass :   TWndClass;     // class struct for fake Winamp window
   FakeWinHandle : HWND = 0; // handle of fake Winamp window
   PlayerMode : TPlayerMode = plmStandby;
   StreamInfo : TStreamInfo;
   SongPosition : DWORD = 0;
   latencyMs, delayMs : DWORD;
   WaitingRendering : boolean = false;

function NewWindowProc(WindowHandle: hWnd; TheMessage: DWORD;
  ParamW: WParam; ParamL: LParam): longInt; stdcall;
var
   BaseWindowHandle, WinHwnd : HWND;
   ProcId : DWORD;
begin
  // Following message handling is to set up synchronized action between the windows of
  // main program and the one of visualizer for the case that visualizer gets
  // focus, hides or shows.
  // (visualizer = Winamp visualization plug-in or its visible window)

  // Put the windows of main program just below the visualizer when visualizer gets focus.
   if (TheMessage = WM_ACTIVATEAPP) then
   begin
      if (ParamW <> 0) and SyncWithMain then
      begin
         PopUpMainWindow;
         NewWindowProc := 0;  // Exit here
         exit;
      end;
   end else

 // Restore(=minimized state to normal state) the windows of main program when visualizer
 // restores.
   if (TheMessage = WM_SIZE) then
   begin
      if (ParamW = SIZE_RESTORED) and SyncWithMain  then
      begin
         WinHwnd := GetWindow(MainWindowHandle, GW_HWNDFIRST);

      // main program may have several visible windows so investigate every visible
      // windows and restores it if its owner is main program.
         repeat
            BaseWindowHandle := WinHwnd;
            WinHwnd := GetWindow(BaseWindowHandle, GW_HWNDNEXT);
            GetWindowThreadProcessId(WinHwnd, @ProcId);
            if ProcId = MainProcId then
               if IsWindowVisible(WinHwnd) then
                  if IsIconic(WinHwnd) then
                     ShowWindow(WinHwnd, SW_RESTORE);
         until (WinHwnd = 0);

      end;
   end else

 // Terminate VisOut program if visualizer window is closed
   if (TheMessage = WM_DESTROY) then
   begin
      if not EndByProgram then
      begin
         VisualizerQuitted := true;
         SetVisualizerQuitted;   // set PluginCtrl unit's "VisualizerQuitted"

       //   Application.Terminate;  // <- may not end some plug-ins by this sentence
                                    // So, I use PostMessage instead
         PostMessage(VisOutHandle, DataReadyMsg, QuitProg, 0);
         NewWindowProc := 0;  // Exit here
         exit;
      end;
   end;

 // Call the original Window procedure
   NewWindowProc := CallWindowProc(OldWindowProc, WindowHandle, TheMessage,
                                         ParamW, ParamL);
end;

procedure PopUpMainWindow;
var
   WinHwnd : HWND;
   BaseWindowHandle : HWND;
   ProcId : DWORD;
   ClassName : array[0..255] of Char;
begin
 // get the handle of the window laid just below visualizer
   BaseWindowHandle := VisWindowHandle;
   repeat
      WinHwnd := GetWindow(BaseWindowHandle, GW_HWNDNEXT);
      BaseWindowHandle := WinHwnd;             // for next iteration
      if WinHwnd <> 0 then
         if IsWindowVisible(WinHwnd) then
            if not IsIconic(WinHwnd) then
            begin
               GetClassName(WinHwnd, ClassName, SizeOf(ClassName));
               if string(ClassName) <> 'TApplication' then
                  break;
            end;
   until (WinHwnd = 0);

   if WinHwnd <> 0 then  // found a window
   begin
      GetWindowThreadProcessId(WinHwnd, @ProcId); // get the process Id.
      if ProcId = MainProcId then   // Is it Main program's process Id. ?
         exit;
   end else
      exit;


  // get the handle of a main program's window
   repeat
      BaseWindowHandle := WinHwnd;    // for next iteration
      WinHwnd := GetWindow(BaseWindowHandle, GW_HWNDNEXT);
      if WinHwnd <> 0 then
         if IsWindowVisible(WinHwnd) then
            if not IsIconic(WinHwnd) then
            begin
               GetClassName(WinHwnd, ClassName, SizeOf(ClassName));
               if string(ClassName) <> 'TApplication' then
               begin
                  GetWindowThreadProcessId(WinHwnd, @ProcId);
                  if ProcId = MainProcId then  // Is it main program's process Id. ?
                     break;
               end;
            end;
   until (WinHwnd = 0);

   if WinHwnd = 0 then
      exit;

 // Put the windows of main program just below the visualizer
   SetWindowPos(WinHwnd, VisWindowHandle, 0, 0, 0, 0,
                                          SWP_NOACTIVATE+SWP_NOMOVE+SWP_NOSIZE);
end;


procedure TdumbForm.RenderToVis;
var
   p1 : PDWORD;
   p2 : PBYTE;
   i : integer;
   RenderResult : integer;
begin
   if VisualizerQuitted then
      exit;

   p1 := ShareMemPointer;
   inc(p1, 280); // FFT Data region is allocated from offset 280 thru 2595 in shared memory
   inc(p1, 1);   // p1^ : flag to notify that new data is given
   Vismod[VisModNum]^.sRate := p1^;
   inc(p1, 1);
   Vismod[VisModNum]^.nCh := p1^;
   inc(p1, 1);
   SongPosition := p1^;
   inc(p1, 1);

   p2 := pointer(p1);
   for i := 1 To 576 do
   begin
     Vismod[VisModNum]^.SpectrumData[1, i] := p2^;
     inc(p2, 1);
   end;
   for i := 1 To 576 do
   begin
     Vismod[VisModNum]^.SpectrumData[2, i] := p2^;
     inc(p2, 1);
   end;

   for i := 1 to 576 do
   begin
     Vismod[VisModNum]^.Waveformdata[1, i] := p2^;
     inc(p2, 1);
   end;
   for i := 1 To 576 do
   begin
     Vismod[VisModNum]^.Waveformdata[2, i] := p2^;
     inc(p2, 1);
   end;

  // Some plug-ins may have bugs which cause exceptions during rendering operation.
  // Visout will terminate if it encounters an exception.
  // My system have crashed if I use the exception handling routine to avoid
  // abnornal termination. ( I have tested with vis_jnp.dll )

  // try
     RenderResult := Vismod[VisModNum].Render(Vismod[VisModNum]);
  { except
   // Neglect some kind of exceptions
     on E: EMathError do RenderResult := 0;
     else
        Application.HandleException(Self);
   end; }

   if RenderResult <> 0 then
   begin
      Vismod[VisModNum].Quit(Vismod[VisModNum]);
      VisualizerQuitted := true;
      SetVisualizerQuitted;
   end;

   p1 := ShareMemPointer;
   inc(p1, 280);
   p1^ := 0;  // Notify that rendering is completed
end;

procedure TdumbForm.EndProgram;
begin
   EndByProgram := true;
   if NumVismod > 0 then
      UnloadVisModule;
   NumVismod := 0;

   Close;
end;

procedure TdumbForm.SetSyncWithMain;
var
   p1 : PDWORD;
begin
   p1 := ShareMemPointer;
   if p1^ = 0 then
      SyncWithMain := false
   else
      SyncWithMain := true;
end;

procedure TdumbForm.HideVisWindow;
begin
   CloseWindow(VisWindowHandle);
end;


procedure TdumbForm.ShowVisWindow;
begin
   if IsIconic(VisWindowHandle) then
      OpenIcon(VisWindowHandle);
end;

procedure TdumbForm.GetStatus(StatusId : DWORD);
var
   p1 : PDWORD;
   p2 : PBYTE;
   ModeNum : DWORD;
   TitleBuf : array[0..255] of char;
begin
   p1 := ShareMemPointer;

   if StatusId = InformPlayerMode then
   begin
      ModeNum := p1^;
      case ModeNum of
         0 : PlayerMode := plmStandby;
         1 : PlayerMode := plmReady;
         2 : PlayerMode := plmStopped;
         3 : PlayerMode := plmPlaying;
         4 : PlayerMode := plmPaused;
         else
            PlayerMode := plmStandby;
      end;
   end else if StatusId = InformStreamInfo then
   begin
      inc(p1, 4);
      StreamInfo.SampleRate := p1^;
      inc(p1, 1);
      StreamInfo.BitRate := p1^;
      inc(p1, 1);
      StreamInfo.Channels := p1^;
      inc(p1, 1);
      StreamInfo.Duration := p1^;
      inc(p1, 1);
      p2 := pointer(p1);
      move(p2^, TitleBuf, 256);
      StreamInfo.Title := StrPas(TitleBuf);
      SetWindowText(FakeWinHandle, pchar(StreamInfo.Title + ' - Winamp'));
   end else if StatusId = InformSyncWindows then
   begin
      inc(p1, 276);
      if p1^ = 0 then
         SyncWithMain := false
      else
         SyncWithMain := true;
   end;
end;


procedure TdumbForm.HandleMessages(var Msg: TMsg; var Handled: Boolean);
begin
   if (Msg.Message = DataReadyMsg) then
   begin
      case Msg.wParam of
         DataReady : begin
                        if WaitingRendering then
                        begin
                           Application.ProcessMessages;
                           exit;
                        end;
                        WaitingRendering := true;
                        RenderToVis;
                        WaitingRendering := false;
                     end;

         QuitProg : EndProgram;
         SyncWindows : SetSyncWithMain;
         MinimizeWindow : HideVisWindow;
         RestoreWindow : ShowVisWindow;
         InformPlayerMode..InformSyncWindows : GetStatus(Msg.wParam);
      end;
      Handled := True;
   end;
end;

procedure TdumbForm.FormCreate(Sender: TObject);
var
   Visheader : PWinampVisHeader;
   ErrCode : integer;
   Msg : pchar;
   PlayerModeId : DWORD;

begin
   if not fisSharedMemory1.OpenMemory then
   begin
      Msg := 'Shared memory is not available' + chr(10) +
             '(This program is designed to be called by TBassPlayer)';
      MessageBox(Application.Handle, Msg, 'Error', MB_OK + MB_ICONERROR);
      Application.Terminate;
      exit;
   end;

   DataReadyMsg := fisSharedMemory1.SendingDataMsg;
   ShareMemPointer := fisSharedMemory1.MapMemory;
   GetWindowThreadProcessId(Self.Handle, @VisProcId);
   Application.OnMessage := HandleMessages;

   PluginName := ParamStr(1);
   VisModNum := strToint(ParamStr(2));
   MainProcId := strToint64(ParamStr(3));
   MainWindowHandle := strToint64(ParamStr(4));
   MsgHandle := strToint64(ParamStr(5));
   StreamInfo.Title := ParamStr(6);
   StreamInfo.Channels := strToint(ParamStr(7));
   StreamInfo.SampleRate := strToint(ParamStr(8));
   StreamInfo.BitRate := strToint(ParamStr(9));
   StreamInfo.Duration := strToint(ParamStr(10));

   PlayerModeId := strToint(ParamStr(11));
   case PlayerModeId of
       0 : PlayerMode := plmStandby;
       1 : PlayerMode := plmReady;
       2 : PlayerMode := plmStopped;
       3 : PlayerMode := plmPlaying;
       4 : PlayerMode := plmPaused;
   end;

   LoadVisModule(PluginName, Visheader, Vismod, NumVismod, FakeWinHandle);
   if (NumVismod > 0) and (VisModNum < NumVismod) and (VisModNum >= 0) then
   begin
      ErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
      Vismod[VisModNum]^.sRate := StreamInfo.SampleRate;
      Vismod[VisModNum]^.nCh := StreamInfo.Channels;
      ErrCode := StartVisModule(VisModNum);
      if ErrCode = 0 then
      begin
         VisOutHandle := self.Handle;
         EndByProgram := false;
         if strToint(ParamStr(12)) = 0 then
            SyncWithMain := false
         else
            SyncWithMain := true;

         delayMs := Vismod[VisModNum]^.delayMs;
         latencyMs := Vismod[VisModNum]^.latencyMs;
         PostMessage(MsgHandle, DataReadyMsg, VisModuleLoaded, (latencyMs shl 16) + delayMS);
         Timer1.Enabled := true;
      end else
      begin
         UnloadVisModule;
         Application.MessageBox(pchar('Specified plug-in is not operative, Error Code = '
                               + intToStr(ErrCode)), 'Error', MB_OK + MB_ICONERROR);
         PostMessage(MsgHandle, DataReadyMsg, StartVisOut, 0);
         Application.Terminate;
      end;
   end else begin
      if NumVismod = 0 then
         Msg := 'Specified plug-in is invalid'
      else
         Msg := 'Invalid module number';
      Application.MessageBox(Msg, 'Error', MB_OK + MB_ICONERROR);
      PostMessage(MsgHandle, DataReadyMsg, StartVisOut, 0);
      Application.Terminate;
   end;
end;

procedure TdumbForm.FormDestroy(Sender: TObject);
begin
   SetErrorMode(ErrorMode);

 // inform main program that visualizer is ended
   PostMessage(MsgHandle, DataReadyMsg, EndVisOut, 0);

   if ShareMemPointer <> nil then
     fisSharedMemory1.UnMapMemory(ShareMemPointer);
   fisSharedMemory1.CloseMemory;
end;


function LookAtAllWindows(Handle: HWnd; Temp: Longint): BOOL; stdcall;
var
  ProcId : DWORD;
begin
   result := true;

   if IsWindowVisible(Handle) then
   begin
      GetWindowThreadProcessId(Handle, @ProcId); // get process identifier
      if ProcId = VisProcId then
         if Handle <> FakeWinHandle then
         begin
            VisWindowHandle := Handle;
            result := false;
         end;
   end;
end;

procedure TdumbForm.Timer1Timer(Sender: TObject);
begin
   Timer1.Enabled := false;

   EnumWindows(@LookAtAllWindows, 0);
   if VisWindowHandle <> 0 then
   begin
      OldWindowProc := Pointer(SetWindowLong(VisWindowHandle, GWL_WNDPROC,
                                             LongInt(@NewWindowProc)));
      SetVisOutHandle(VisWindowHandle);

    // inform main program that visualizer is activated
      PostMessage(MsgHandle, DataReadyMsg, StartVisOut, VisWindowHandle);

    // Following sentence is to display the title of playing stream file on
    // visualizer's display.
    // You must add ' - Winamp' to the title string because some visualization plug-ins
    // show erroneous operation if it is missed (ex. vis_Bass-C.dll)
      SetWindowText(FakeWinHandle, pchar(StreamInfo.Title + ' - Winamp'));
   end else
      Timer1.Enabled := true;  // repeat until visualizer's window is detected
end;



// ----------------------------------------------------------------------------
// Following codes are for emulation of Winamp's main window

var
   hInst : HWND;         // handle of program (hinstance)

const
   WM_WA_IPC = WM_USER;
   IPC_GETVERSION = 0;
   IPC_ISPLAYING = 104;
   IPC_GETOUTPUTTIME = 105;
   IPC_WRITEPLAYLIST =120;
   IPC_GETINFO = 126;


// This WindowProc function processes some essential Winamp IPC messages sent from
// visualization plug-in to fake Winamp window, to deceive visualization plug-ins
// as if the cooperator of them(= visualization plug-ins) is Winamp ver 2.4.
// See Winamp_IPC.txt for Winamp IPC messages.
function WindowProc(hWnd, Msg, wParam, lParam : Longint) : Longint; stdcall;
begin
   if Msg = WM_WA_IPC then
   begin
      if LParam = IPC_GETVERSION then
         Result := $2040    // acts as if the main program is Winamp ver 2.4
      else if LParam = IPC_ISPLAYING then
      begin
         if PlayerMode = plmPlaying then
            Result := 1
         else if PlayerMode = plmPaused then
            Result := 3
         else
            Result := 0;
      end else if LParam = IPC_GETINFO then
      begin
         if WParam = 0 then           // Sample rate
            Result := StreamInfo.SampleRate
         else if WParam = 1 then      // Bit rate
            Result := StreamInfo.BitRate
         else if WParam = 2 then      // Channels
            Result := StreamInfo.Channels;
      end
      else if LParam = IPC_GETOUTPUTTIME then
      begin
         if PlayerMode <> plmPlaying then
            Result := -1
         else if WParam = 0 then      // position in miliseconds
            Result := SongPosition
         else if WParam = 1 then      // song length in seconds
            Result := StreamInfo.Duration div 1000
      end
   end
   else
      Result := DefWindowProc(hWnd, Msg, wParam, lParam);

end;

// Create a fake window which acts like the one of Winamp's main window
function CreateFakeWindow : HWND;      // * New at Ver 1.2
var
   WinAtom : TAtom;
begin
   if FakeWinHandle <> 0 then   // Avoid duplicate creation of fake Winamp window
   begin
      result := FakeWinHandle;
      exit;
   end;

  hInst := GetModuleHandle(nil); // get the application instance

  with wClass do
  begin
    Style         := CS_PARENTDC;
    hIcon         := LoadIcon(hInst, 'MAINICON');
    lpfnWndProc   := @WindowProc;
    hInstance     := hInst;
    hbrBackground := COLOR_BTNFACE + 1;
    lpszClassName := 'Winamp v1.x';
    hCursor       := LoadCursor(0, IDC_ARROW);
  end;

  // Once our class is registered we can start making windows with it
  WinAtom := windows.RegisterClass(wClass);
  if WinAtom <> 0 then
     FakeWinHandle := CreateWindow('Winamp v1.x', 'Winamp 2.40',
                                   WS_POPUP,      // no-frame, non-visible window
                                   5, 5, 25, 25,  // x, y, width, height
                                   0, 0, hInst, nil)
  else
     FakeWinHandle := 0;

  result := FakeWinHandle;
end;

function DestroyFakeWindow : boolean;  // * New at Ver 1.2
begin
   if FakeWinHandle <> 0 then
   begin
      result := DestroyWindow(FakeWinHandle); 	// handle to window to destroy
      if result then
      begin
         FakeWinHandle := 0;
         windows.UnRegisterClass('Winamp v1.x', hInst);
      end;
   end else
      result := false;
end;


end.
