{ Original Source : BASSWMA.PAS by Ian Luck
  BASSWMA 2.0 Delphi API, copyright (c) 2002-2003 Ian Luck.
  Requires BASS 2.0 - available from www.un4seen.com

  See the BASSWMA.CHM file for more complete documentation
}

Unit RT_BASSWMA;

interface

uses windows, Dynamic_bass;

const
  // Additional error codes returned by BASS_ErrorGetCode
  BASS_ERROR_WMA_LICENSE     = 1000; // the file is protected
  BASS_ERROR_WMA_WM9         = 1001; // WM9 is required
  BASS_ERROR_WMA_DENIED      = 1002; // access denied (user/pass is invalid)

  // Additional flags for use with BASS_WMA_EncodeOpen/File/Network/Publish
  BASS_WMA_ENCODE_TAGS       = $10000; // set tags in the WMA encoding
  BASS_WMA_ENCODE_SCRIPT     = $20000; // set script (mid-stream tags) in the WMA encoding

  // Additional flag for use with BASS_WMA_EncodeGetRates
  BASS_WMA_ENCODE_RATES_VBR  = $10000; // get available VBR quality settings

  // WMENCODEPROC "type" values
  BASS_WMA_ENCODE_HEAD = 0;
  BASS_WMA_ENCODE_DATA = 1;
  BASS_WMA_ENCODE_DONE = 2;

  // BASS_CHANNELINFO type
  BASS_CTYPE_STREAM_WMA      = $10300;


type
  HWMENCODE = DWORD;		// WMA encoding handle

  CLIENTCONNECTPROC = procedure(handle:HWMENCODE; connect:BOOL; ip:PChar; user:DWORD); stdcall;
  {
    Client connection notification callback function.
    handle : The encoder
    connect: TRUE=client is connecting, FALSE=disconnecting
    ip     : The client's IP (xxx.xxx.xxx.xxx:port)
    user   : The 'user' parameter value given when calling BASS_EncodeSetNotify */
  }

  WMENCODEPROC = procedure(handle:HWMENCODE; dtype:DWORD; buffer:Pointer; length:DWORD; user:DWORD); stdcall;
  {
    Encoder callback function.
    handle : The encoder handle
    dtype  : The type of data, one of BASS_WMA_ENCODE_xxx values
    buffer : The encoded data
    length : Length of the data
    user   : The 'user' parameter value given when calling BASS_WMA_EncodeOpen
  }

var
  BASS_WMA_StreamCreateFile : function(mem:BOOL; fl:pointer; offset,length,flags:DWORD): HSTREAM; stdcall;
  BASS_WMA_GetIWMReader : function(handle:HSTREAM): Pointer; stdcall;

  BASS_WMA_EncodeGetRates : function(freq,flags:DWORD): Pointer; stdcall;
  BASS_WMA_EncodeOpen : function(freq,flags,bitrate:DWORD; fname:PChar): HWMENCODE; stdcall;
  BASS_WMA_EncodeOpenFile : function(freq,flags,bitrate:DWORD; fname:PChar): HWMENCODE; stdcall;
  BASS_WMA_EncodeOpenNetwork : function(freq,flags,bitrate,port,clients:DWORD): HWMENCODE; stdcall;
  BASS_WMA_EncodeOpenPublish : function(freq,flags,bitrate:DWORD; fname:PChar): HWMENCODE; stdcall;
  BASS_WMA_EncodeGetPort : function(handle:HWMENCODE): DWORD; stdcall;
  BASS_WMA_EncodeSetNotify : function(handle:HWMENCODE; proc:CLIENTCONNECTPROC; user:DWORD): BOOL; stdcall;
  BASS_WMA_EncodeGetClients : function(handle:HWMENCODE): DWORD; stdcall;
  BASS_WMA_EncodeSetTag : function(handle:HWMENCODE; tag,text:PChar): BOOL; stdcall;
  BASS_WMA_EncodeWrite : function(handle:HWMENCODE; buffer:Pointer; length:DWORD): BOOL; stdcall;
  BASS_WMA_EncodeClose : procedure(handle:HWMENCODE); stdcall;


var BASSWMA_Handle : Thandle = 0;

function Load_BASSWMADLL(const dllfilename : string) : boolean;
procedure Unload_BASSWMADLL;

implementation

function Load_BASSWMADLL(const dllfilename : string) : boolean;
var
   oldmode : integer;
begin
   if BASSWMA_Handle <> 0 then // is it already there ?
      result := true
   else begin {go & load the dll}
   oldmode := SetErrorMode($8001);
   BASSWMA_Handle := LoadLibrary(pchar(dllfilename));  // obtain the handle we want
   SetErrorMode(oldmode);
   if BASSWMA_Handle <> 0 then
       begin {now we tie the functions to the VARs from above}

   @BASS_WMA_StreamCreateFile := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_StreamCreateFile');
   @BASS_WMA_GetIWMReader := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_GetIWMReader');
   @BASS_WMA_EncodeGetRates := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeGetRates');
   @BASS_WMA_EncodeOpen := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeOpen');
   @BASS_WMA_EncodeOpenFile := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeOpenFile');
   @BASS_WMA_EncodeOpenNetwork := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeOpenNetwork');
   @BASS_WMA_EncodeOpenPublish := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeOpenPublish');
   @BASS_WMA_EncodeGetPort := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeGetPort');
   @BASS_WMA_EncodeSetNotify := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeSetNotify');
   @BASS_WMA_EncodeGetClients := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeGetClients');
   @BASS_WMA_EncodeSetTag := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeSetTag');
   @BASS_WMA_EncodeWrite := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeWrite');
   @BASS_WMA_EncodeClose := GetProcAddress(BASSWMA_Handle, 'BASS_WMA_EncodeClose');

 // check if everything is linked in correctly
   if (@BASS_WMA_StreamCreateFile = nil) or
      (@BASS_WMA_GetIWMReader = nil) or
      (@BASS_WMA_EncodeGetRates = nil) or
      (@BASS_WMA_EncodeOpen = nil) or
      (@BASS_WMA_EncodeOpenFile = nil) or
      (@BASS_WMA_EncodeOpenNetwork = nil) or
      (@BASS_WMA_EncodeOpenPublish = nil) or
      (@BASS_WMA_EncodeGetPort = nil) or
      (@BASS_WMA_EncodeSetNotify = nil) or
      (@BASS_WMA_EncodeGetClients = nil) or
      (@BASS_WMA_EncodeSetTag = nil) or
      (@BASS_WMA_EncodeWrite = nil) or
      (@BASS_WMA_EncodeClose = nil) then
       begin
          FreeLibrary(BASSWMA_Handle);
          BASSWMA_Handle := 0;
       end;
     end;
     result := (BASSWMA_Handle <> 0);
   end;
end;

procedure Unload_BASSWMADLL;
begin
   if BASSWMA_Handle <> 0 then
      FreeLibrary(BASSWMA_Handle);

   BASSWMA_Handle := 0;
end;

end.
